# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import sys
import os
import tempfile
import logging
import subprocess
import shlex
import shutil
import math
from math import *  # noqa: F401,F403
from typing import Any, Callable, Dict, List, Iterable, Union


# =============================================================================
# XPR Eval helper functions.

def register_eval_function_stack() -> Dict[str, Callable]:
    """
    Returns the basic stack of evaluable Functions.
    """
    globals_ = globals()

    function_dict = dict([
        (f_name, globals_.get(f_name, None)) for f_name in dir(math) if not f_name.startswith('_')
    ])
    function_dict['str'] = str
    function_dict['len'] = len

    def _switch_func(value, options, default_val=None):
        """
        Select one of many codes to return it related value.
        """
        result = list(filter(lambda option: option[0] == value, options))
        return result[0][1] if len(result) > 0 else default_val

    function_dict['switch'] = _switch_func
    function_dict['select'] = _switch_func

    def _iif_func(condition, true_value, false_value):
        """
        Return true/false value according condition.
        """
        return true_value if condition else false_value

    function_dict['iif'] = _iif_func

    return function_dict


# Stack of callable math functions to include in an eval() execution, only MATH functions for safety.
BASIC_EVAL_FUNCTION_STACK = None

# =============================================================================


class ProcessingUtils:
    """
    Provides generic utility functions for Processing tasks.
    """
    @staticmethod
    def cast_value(value, value_type):
        """
        Cast a value to the specified type.
        """
        if value_type == float:
            return float(value)
        if value_type == int:
            return int(value)
        if value_type == list:
            return list(value)

        return value

    @staticmethod
    def cast_enum(value, class_enum_type):
        """
        Cast a value to the specified Enum-like type.
        """
        return value if isinstance(value, int) else class_enum_type().__getattribute__(value)

    @staticmethod
    def object_as_dict(obj: object) -> Dict[str, Any]:
        """
        Converts the specified Object to Dict.
        """
        new_dict = dict((k, v) for k, v in obj.__dict__.items() if not callable(v) and not k.startswith('__'))
        return new_dict

    @staticmethod
    def strtobool(value) -> bool:
        """
        Converts to bool type the specified string value.
        """
        if type(value) == str:
            value = value.lower()

            if value == 'false' or value == '0' or value == 'no':
                return False
            if value == 'true' or value == '1' or value == 'yes':
                return True

        return value

    @staticmethod
    def eval_function(expression, attributes: Union[List, Dict] = {}, normalize_attribs: bool = True):
        """
        Evaluate the specified expression and arguments.
        """
        if isinstance(expression, str) and \
           expression.find('$') == -1 and (os.path.isdir(expression) or os.path.isfile(expression)):
            return expression

        if normalize_attribs:
            attribute_dict = dict()

            if type(attributes) is list:
                for temp_dict in attributes:
                    attribute_dict.update(('_attr__'+str(k), v) for k, v in temp_dict.items())
            else:
                attribute_dict.update(('_attr__'+str(k), v) for k, v in attributes.items())
        else:
            if type(attributes) is list:
                attribute_dict = dict()
                for temp_dict in attributes:
                    attribute_dict.update(temp_dict)
            else:
                attribute_dict = attributes

        global BASIC_EVAL_FUNCTION_STACK

        if BASIC_EVAL_FUNCTION_STACK is None:
            BASIC_EVAL_FUNCTION_STACK = register_eval_function_stack()

        if isinstance(expression, str):
            xpr = expression
            arr = ['_attr__' if c == '$' and xpr[i+1].isalpha() else c for i, c in enumerate(xpr)]
            xpr = ''.join(arr)
            xpr = xpr.replace("\\", "^^")
            # r = expression.replace('$', '_attr__')
            res = eval(xpr, BASIC_EVAL_FUNCTION_STACK, attribute_dict)
            if isinstance(res, str):
                res = res.replace("^^", "\\")

            return res
        else:
            return eval(expression, BASIC_EVAL_FUNCTION_STACK, attribute_dict)


class ProcessingArgs:
    """
    Provides utilities for a Processing task.

    Args:
        temp_path: Path where temporary files are saved.
    """
    def __init__(self, temp_path: str = None):
        self._is_windows = sys.platform.startswith('win')
        self._temp_path = temp_path
        self._temp_data_path = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.dispose()

    def __del__(self):
        self.dispose()

    @staticmethod
    def _normalize_path(path: str) -> str:
        """
        Returns the normalized version of the specified path.
        """
        if ' ' in path and '"' not in path:
            path = '"' + path + '"'
        if '*' in path and '"' not in path:
            path = '"' + path + '"'
        return path

    @staticmethod
    def _get_command_line_args(command_args: Iterable[str]) -> str:
        """
        Returns the command-line string of the specified command_args
        """
        command_line = ''

        if not command_args:
            return command_line
        for command_arg in command_args:
            command_line += ' ' + ProcessingArgs._normalize_path(command_arg)

        return command_line

    @staticmethod
    def _read_subprocess_stdout(p) -> Iterable[str]:
        """
        Prompt the stdout pipeline as real-time style
        """
        while p.poll() is None:
            for line in iter(lambda: p.stdout.readline() if p.poll() is None else '', ''):
                line = line.decode('utf-8', errors='ignore').rstrip()
                if line:
                    yield line

    def temp_data_path(self) -> str:
        """
        Returns an unique temporal data path for this processing task.
        """
        if self._temp_data_path:
            return self._temp_data_path
        elif not self._temp_path:
            self._temp_data_path = tempfile.mkdtemp(prefix='pyproc-task_')
            return self._temp_data_path
        else:
            self._temp_data_path = self._temp_path

            # 'temp_path' is only created and removed when disposing if that folder does not exist.
            if not os.path.exists(self._temp_path):
                self._temp_path = None

            os.makedirs(self._temp_data_path, exist_ok=True)
            return self._temp_data_path

    def dispose(self) -> None:
        """
        Dispose all resources including temporary data folder.
        """
        if self._temp_data_path and not self._temp_path:
            shutil.rmtree(self._temp_data_path, ignore_errors=True)

    def run(self, command_line: Union[str, Iterable[str]], environment_args: Dict[str, str] = None) -> int:
        """
        Runs the specified command.
        """
        if not type(command_line) is str:
            command_line = ProcessingArgs._get_command_line_args(command_line)

        env_args = os.environ.copy()
        if environment_args:
            env_args.update(environment_args)

        logging.debug('command_line=' + command_line)
        logging.debug('command_args=' + str(shlex.split(command_line, posix="win" not in sys.platform)))

        try:
            with subprocess.Popen(command_line,
                                  shell=True,
                                  stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env_args) as p:
                if p.stderr:
                    message = ''

                    for line in p.stderr.readlines():
                        message += line.decode('utf-8').rstrip() + '\n'

                    message = message.rstrip()
                    logging.error(message)
                    raise RuntimeError(message)

                if p.stdout:
                    for message in ProcessingArgs._read_subprocess_stdout(p):
                        logging.info(message)

                return p.returncode

        except subprocess.CalledProcessError as e:
            raise e



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from typing import Dict
from uuid import uuid4
from collections.abc import Iterable


class ModuleIt(Iterable):
    """
    Wrapper of a Module to run it as Iterator.
    """
    def __init__(self, object_it, module, pipeline_args, processing_args):
        Iterable.__init__(self)
        self._object_it = object_it
        self.module = module
        self.pipeline_args = pipeline_args
        self.processing_args = processing_args

    def __del__(self):
        self._object_it = None
        self.module = None
        self.pipeline_args = None
        self.processing_args = None

    def __next__(self):
        return next(self._object_it)

    def __iter__(self):
        return self


class AbstractModule(Iterable):
    """
    Generic Module that operates on Workflows of Geospatial data.
    """
    def __init__(self):
        Iterable.__init__(self)
        self.className = str(self.__class__.__name__)
        self.classType = 'module'
        self.stageId = str(uuid4())
        # logging.debug('-> New object {} created!'.format(self.className))

    def __del__(self):
        # logging.debug('<- {} destroyed!'.format(self.className))
        pass

    def __iter__(self) -> ModuleIt:
        """
        Returns the iterable set of Geospatial features of current Workflow.
        """
        if self.pipeline_args is None:
            return iter([])

        pipeline_args = self.pipeline_args
        processing_args = pipeline_args.processing_args
        return ModuleIt(
            self.run(pipeline_args.data_source, processing_args), self, pipeline_args, processing_args
        )

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return self.className

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Modules'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Generic module'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {}

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        return schema_def

    def run(self, data_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        if data_store is not None:
            for row in data_store:
                yield row

    def finished_run(self, pipeline, processing_args):
        """
        Finishing a Workflow on Geospatial data.
        """
        return False



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import logging
import re
from typing import Any, Callable, Dict, Iterable, List, Tuple, Union

import pyproj as pj
from shapely.geometry.polygon import Polygon
from shapely.geometry.base import BaseGeometry
from shapely.geos import lgeos
from shapely.ops import transform

from geodataflow.core.capabilities import StoreCapabilities


class DataUtils:
    """
    Provides generic Data/File utility functions.
    """
    @staticmethod
    def get_layer_name(connection_string: str) -> str:
        """
        Returns the FeatureClass name of the specified GDAL/OGR ConnectionString.
        """
        if DataUtils.represents_json_feature_collection(connection_string):
            return 'FeatureCollection'

        file_name, file_ext = os.path.splitext(connection_string)
        if file_name and file_ext:
            tmp_key = os.path.basename(file_name)
            tmp_pos = tmp_key.find('.')
            return tmp_key[:tmp_pos] if tmp_pos != -1 else tmp_key

        raise Exception('Unknown LayerName parsing of the ConnectionString="{}"'.format(connection_string))

    @staticmethod
    def enumerate_single_connection_string(connection_string: Union[str, Iterable[str]]) -> Iterable[str]:
        """
        Enumerates the single connection string from the specified connection string.
        """
        if DataUtils.represents_json_feature_collection(connection_string):
            yield connection_string
            return

        if isinstance(connection_string, str):
            for connection_item in re.split(';|\n', connection_string):
                if '*' in connection_item:
                    import glob
                    for file_name in glob.glob(connection_item):
                        yield file_name
                else:
                    yield connection_item
            return

        for connection_item in connection_string:
            for temp_c in DataUtils.enumerate_single_connection_string(connection_item):
                yield temp_c

        pass

    @staticmethod
    def represents_json_feature_collection(connection_string: Union[str, List[str], Dict[str, Any]]) -> bool:
        """
        Returns if the specified connection string is a GeoJSON Feature collection.
        """
        if isinstance(connection_string, list) and connection_string:
            connection_string = connection_string[0]
        if isinstance(connection_string, dict) and connection_string.get('type', '') == 'FeatureCollection':
            return True
        if isinstance(connection_string, str) and len(connection_string) > 32:
            temp_text = connection_string.replace("\n", "").replace(" ", "")[0:32]

            if temp_text.startswith("{'type':'FeatureCollection'"):
                return True
            if temp_text.startswith('{"type":"FeatureCollection"'):
                return True

        return False


class GdalUtils:
    """
    Provides generic GDAL/OGR utility functions.
    """
    @staticmethod
    def get_gdal_driver(connection_string: str):
        """
        Return the GDAL Driver that supports the specified DataSource.
        """
        from geodataflow.geoext.gdalenv import GdalEnv
        gdal = GdalEnv.default().gdal()

        file_name, file_ext = os.path.splitext(connection_string)
        if file_name and file_ext:
            temp_key = os.path.basename(connection_string)
            temp_pos = temp_key.find('.')
            file_ext = temp_key[temp_pos:]

        file_ext = file_ext.lower()

        if file_ext in ['.tiff', '.tif']:
            return gdal.GetDriverByName('Gtiff')
        if file_ext in ['.jpeg', '.jpg']:
            return gdal.GetDriverByName('JPEG')
        if file_ext == '.png':
            return gdal.GetDriverByName('PNG')
        if file_ext == '.vrt':
            return gdal.GetDriverByName('VRT')
        if file_ext == '.gpkg':
            return gdal.GetDriverByName('GPKG')
        if file_ext == '.jp2':
            driver = gdal.GetDriverByName('JP2OpenJPEG')
            driver = driver if driver else gdal.GetDriverByName('JP2ECW')
            return driver

        raise Exception('Unknown GDAL Driver for the ConnectionString="{}"'.format(connection_string))

    @staticmethod
    def get_ogr_driver(connection_string: str):
        """
        Return the OGR Driver that supports the specified DataSource.
        """
        from geodataflow.geoext.gdalenv import GdalEnv
        ogr = GdalEnv.default().ogr()

        if DataUtils.represents_json_feature_collection(connection_string):
            return ogr.GetDriverByName('GeoJSON')
        if connection_string.startswith('PG:'):
            return ogr.GetDriverByName('PostgreSQL')

        file_name, file_ext = os.path.splitext(connection_string)
        if file_name and file_ext:
            temp_key = os.path.basename(connection_string)
            temp_pos = temp_key.find('.')
            file_ext = temp_key[temp_pos:]

        file_ext = file_ext.lower()

        if file_ext == '.gpkg':
            return ogr.GetDriverByName('GPKG')
        if file_ext in ['.shp', '.shz', '.shp.zip']:
            return ogr.GetDriverByName('ESRI Shapefile')
        if file_ext in ['.json', '.geojson']:
            return ogr.GetDriverByName('GeoJSON')
        if file_ext in ['.kml', '.kmz']:
            return ogr.GetDriverByName('KML')

        raise Exception('Unknown OGR Driver for the ConnectionString="{}"'.format(connection_string))

    @staticmethod
    def test_gdal_capability(connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns if any GDAL Driver supports the specified named Capability.
        """
        try:
            driver = GdalUtils.get_gdal_driver(connection_string)
            if driver is None:
                return False

            from geodataflow.geoext.gdalenv import GdalEnv
            gdal = GdalEnv.default().gdal()

            if capability == StoreCapabilities.READ:
                metadata = driver.GetMetadata()
                is_ok = gdal.DCAP_OPEN in metadata and metadata[gdal.DCAP_OPEN] == 'YES'
                return is_ok
            if capability == StoreCapabilities.WRITE or capability == StoreCapabilities.CREATE:
                metadata = driver.GetMetadata()
                is_ok = (gdal.DCAP_CREATE in metadata and metadata[gdal.DCAP_CREATE] == 'YES') or \
                        (gdal.DCAP_CREATECOPY in metadata and metadata[gdal.DCAP_CREATECOPY] == 'YES')
                return is_ok

            return False
        except Exception as e:
            logging.warning(
                'Error getting the GDAL Driver of the ConnectionString "{}", Cause="{}".'
                .format(connection_string, str(e)))

            return False

    @staticmethod
    def test_ogr_capability(connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns if any OGR Driver supports the specified named Capability.
        """
        try:
            driver = GdalUtils.get_ogr_driver(connection_string)
            if driver is None:
                return False

            from geodataflow.geoext.gdalenv import GdalEnv
            ogr = GdalEnv.default().ogr()

            if capability == StoreCapabilities.READ:
                return True
            if capability == StoreCapabilities.WRITE:
                is_ok = driver.TestCapability(ogr.ODrCCreateDataSource)
                return is_ok
            if capability == StoreCapabilities.CREATE:
                is_ok = driver.TestCapability(ogr.ODrCCreateDataSource)
                return is_ok

            return False
        except Exception as e:
            logging.warning(
                'Error getting the OGR Driver of the specified ConnectionString "{}", Cause="{}".'
                .format(connection_string, str(e)))

            return False

    @staticmethod
    def get_ogr_connection_string(connection_string: str) -> str:
        """
        Return the normalized OGR Connection String that supports the specified DataSource.
        """
        if os.path.exists(connection_string):
            file_ext = os.path.splitext(os.path.basename(connection_string))[1]
            file_ext = file_ext.lower()

            if file_ext in ['.kmz', '.zip']:
                from zipfile import ZipFile
                import fnmatch
                import re
                re_ob = re.compile('|'.join(
                    fnmatch.translate('*.' + f_ext) for f_ext in ['kml', 'shp', 'gpkg', 'geojson'])
                )

                with ZipFile(connection_string, mode='r') as zf:
                    for zc in zf.infolist():
                        if re_ob.match(zc.filename):
                            return '/vsizip/' + connection_string + '/' + zc.filename

        return connection_string

    @staticmethod
    def get_spatial_reference(crs_def):
        """
        Returns the OGR SpatialReference from the specified CRS/EPSG/WKT definition.
        """
        from osgeo import osr

        crs = osr.SpatialReference()
        if hasattr(crs, 'SetAxisMappingStrategy'):
            crs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

        if isinstance(crs_def, int):
            crs.ImportFromEPSG(crs_def)
            return crs
        if isinstance(crs_def, str):
            if crs_def.isdigit():
                crs.ImportFromEPSG(int(crs_def))
            elif crs_def.startswith('EPSG:'):
                crs.ImportFromEPSG(int(crs_def[5:]))
            else:
                crs.ImportFromWkt(crs_def)

            return crs

        return crs

    @staticmethod
    def get_spatial_srid(spatial_ref) -> int:
        """
        Returns the EPSG SRID code of the specified SRS SpatialReference.
        """
        if spatial_ref is None:
            return 0

        srid = 0
        try:
            srid = spatial_ref.GetAuthorityCode(None)

            if srid is None or srid == 0 or srid == '':
                temp = spatial_ref.AutoIdentifyEPSG()
                if temp is not None and temp != 0 and temp != '':
                    srid = int(temp)
            else:
                srid = int(srid)

        except Exception as e:
            logging.debug('Error reading the SRID of the CRS="{}", Cause="{}"'.format(str(spatial_ref), e))

        return srid


class GeometryUtils:
    """
    Provides generic CRS/Geometry utility functions.
    """
    @staticmethod
    def get_spatial_crs(crs_def: Union[int, str, pj.CRS]) -> pj.CRS:
        """
        Returns the pyproj CRS from the specified CRS/EPSG/WKT definition.
        """
        if isinstance(crs_def, pj.CRS):
            return crs_def
        if isinstance(crs_def, int):
            return pj.CRS.from_epsg(crs_def)
        if isinstance(crs_def, str) and crs_def.isdigit():
            return pj.CRS.from_epsg(crs_def)

        return pj.CRS.from_string(crs_def)

    @staticmethod
    def create_transform_function(source_crs: Union[int, str, pj.CRS],
                                  target_crs: Union[int, str, pj.CRS]
                                  ) -> Callable[[BaseGeometry], BaseGeometry]:
        """
        Returns a function to transform Geometries between the specified Spatial Reference Systems.
        """
        if not isinstance(source_crs, pj.CRS):
            source_crs = GeometryUtils.get_spatial_crs(source_crs)
        if not isinstance(target_crs, pj.CRS):
            target_crs = GeometryUtils.get_spatial_crs(target_crs)

        if source_crs and target_crs and source_crs.to_epsg() != target_crs.to_epsg():
            transform_fn = pj.Transformer.from_crs(source_crs, target_crs, always_xy=True).transform
            target_srid = target_crs.to_epsg()
        else:
            transform_fn = None
            target_srid = 0

        def transform_fn_(geometry: BaseGeometry) -> BaseGeometry:
            if transform_fn:
                geometry = transform(transform_fn, geometry)
                geometry = geometry.with_srid(target_srid)

            return geometry

        return transform_fn_

    @staticmethod
    def get_srid(obj: Union[int, object, pj.CRS]) -> int:
        """
        Returns the SRID of this Geometry.
        """
        if isinstance(obj, int):
            return obj
        if isinstance(obj, BaseGeometry):
            return lgeos.GEOSGetSRID(obj._geom)
        if isinstance(obj, pj.CRS):
            return obj.to_epsg()
        if hasattr(obj, 'srid'):
            return getattr(obj, 'srid')

        return 0

    @staticmethod
    def set_srid(geometry, obj: Union[int, object, pj.CRS]):
        """
        Set the SRID to specified Geometry.
        """
        srid = GeometryUtils.get_srid(obj)
        if srid:
            if isinstance(geometry, BaseGeometry):
                lgeos.GEOSSetSRID(geometry._geom, srid)
            else:
                setattr(geometry, 'srid', srid)

        return geometry

    @staticmethod
    def create_geometry_from_bbox(x_min: float, y_min: float, x_max: float, y_max: float) -> Polygon:
        """
        Returns the Geometry from the specified BBOX.
        INFO:
            This function adds four mid-point vertices to the Geometry because the transform
            between EPSG:4326 and EPSG:3035 generates curvilinear borders.
        """
        coordinates = [
            (x_min, y_min),
            (x_min + 0.5 * (x_max - x_min), y_min),
            (x_max, y_min),
            (x_max, y_min + 0.5 * (y_max - y_min)),
            (x_max, y_max),
            (x_max - 0.5 * (x_max - x_min), y_max),
            (x_min, y_max),
            (x_min, y_max - 0.5 * (y_max - y_min)),
            (x_min, y_min)
        ]
        return Polygon(coordinates)

    @staticmethod
    def clamp_envelope(envelope: Tuple[float, float, float, float],
                       pixel_size_x: float,
                       pixel_size_y: float,
                       padding_val: int = 0,
                       minimum_ts: int = 0) -> Tuple[float, float, float, float]:
        """
        Clamp the specified Envelope aligned to current GRID.
        """
        x_min = int(envelope[0] / pixel_size_x) * pixel_size_x
        y_min = int(envelope[1] / pixel_size_y) * pixel_size_y
        x_max = int(envelope[2] / pixel_size_x) * pixel_size_x
        y_max = int(envelope[3] / pixel_size_y) * pixel_size_y

        if x_max < envelope[2]:
            x_max += pixel_size_x
        if y_max < envelope[3]:
            y_max += pixel_size_y

        # ... forcing the Envelope have a minimum size of tile/image.
        if minimum_ts:
            num_pixel_x = (x_max - x_min) / pixel_size_x
            num_pixel_y = (y_max - y_min) / pixel_size_y
            pad_pixel_x = num_pixel_x - minimum_ts
            pad_pixel_y = num_pixel_y - minimum_ts

            if pad_pixel_x < 0:
                pad_pixel_x = int(abs(pad_pixel_x) * 0.5)
                x_min = x_min - pixel_size_x * pad_pixel_x
                x_max = x_max + pixel_size_x * (1 + pad_pixel_x)
            if pad_pixel_y < 0:
                pad_pixel_y = int(abs(pad_pixel_y) * 0.5)
                y_min = y_min - pixel_size_y * pad_pixel_y
                y_max = y_max + pixel_size_y * (1 + pad_pixel_y)

        # ... applying a "little" padding ot border of the Envelope.
        if padding_val:
            x_min = x_min - padding_val * pixel_size_x
            y_min = y_min - padding_val * pixel_size_y
            x_max = x_max + padding_val * pixel_size_x
            y_max = y_max + padding_val * pixel_size_y

        return [x_min, y_min, x_max, y_max]


# Inject methods to manage SRID/CRS info in Geometry types.
BaseGeometry.with_srid = GeometryUtils.set_srid
BaseGeometry.get_srid = GeometryUtils.get_srid



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from copy import deepcopy
import os
import sys
import logging
import json
import uuid
import base64
import inspect
from collections import OrderedDict
from typing import Any, Callable, Dict, Iterable, List, Set, Tuple, Union, Type

from geodataflow.core.processingargs import ProcessingUtils
from geodataflow.core.modulemanager import ModuleManager
from geodataflow.core.jsoncomments import JsonComments
from geodataflow.core.schemadef import SchemaDef
from geodataflow.core.capabilities import StoreCapabilities

from geodataflow.pipeline.modules import AbstractModule
from geodataflow.pipeline.basictypes import AbstractReader, AbstractFilter, AbstractWriter


class PipelineModuleManager(ModuleManager):
    """
    Pipeline module manager of dynamic modules deployed in the 'Factories' folder application.
    """
    def __init__(self):
        ModuleManager.__init__(self, (AbstractReader, AbstractFilter, AbstractWriter))

    def find_data_source(self,
                         data_source: str,
                         module_type: Type, capability: StoreCapabilities) -> AbstractModule:
        """
        Returns the first Geospatial DataSource that supports the specified criteria.
        """
        for type_def in set(self.modules.values()):
            if issubclass(type_def, module_type):
                obj = type_def()
                if obj.test_capability(data_source, capability):
                    return obj

        return None


class PipelineManager:
    """
    Provides a framework to process a Workflow of geospatial data.
    Pipelines define the processing of data within a Workflow.
    They describe how data are read, processed and written.
    """
    def __init__(self,
                 config: Dict = {}, custom_modules: Dict[str, Type] = None, custom_modules_path: str = None):
        self.config = config
        self._custom_modules = custom_modules
        self._custom_modules_path = custom_modules_path
        self._pipeline_dir = None
        self._objects: List[AbstractModule] = list()

    @staticmethod
    def _recursive_function(parent_obj,
                            obj,
                            pre_call: bool,
                            function_def: Callable[[Any, Any], None],
                            function_args: Any) -> None:
        """
        Apply the specified function recursively.
        """
        if pre_call:
            function_def(parent_obj, obj, function_args)

        if hasattr(obj, 'ti_'):
            for child in obj.ti_.children:
                PipelineManager._recursive_function(obj, child, pre_call, function_def, function_args)

        if not pre_call:
            function_def(parent_obj, obj, function_args)

        pass

    def objects(self, recursive: bool = False) -> List[AbstractModule]:
        """
        Returns the collection of Operations of this Pipeline.
        """
        if recursive:
            temp_list = list()

            def my_collection_of_objects_function(parent_obj, obj, _):
                temp_list.append(obj)
            for item in self._objects:
                PipelineManager._recursive_function(None, item, True, my_collection_of_objects_function, None)

            return temp_list

        return self._objects

    @staticmethod
    def modules() -> Set:
        """
        Returns the collection of Modules deployed.
        """
        modules_root = os.path.dirname(os.path.abspath(__file__))

        manager = PipelineModuleManager()
        modules = manager.load_modules([
            os.path.join(modules_root, 'readers'),
            os.path.join(modules_root, 'filters'),
            os.path.join(modules_root, 'writers')
        ])
        return set(modules.values())

    @staticmethod
    def catalog() -> Dict[str, Dict[str, Any]]:
        """
        Returns the Catalog of metadata of Modules deployed.
        """
        modules = {}

        for module_def in PipelineManager.modules():
            module_obj = module_def()

            modules[module_obj.className] = {
                'name': module_obj.className,
                'type': module_obj.classType,
                'alias': module_obj.alias(),
                'category': module_obj.category(),
                'description': module_obj.description(),
                'params': module_obj.params()
            }

        return modules

    @staticmethod
    def _get_layer_name(connection_string: Union[str, List[str], Dict[str, Any]]) -> str:
        """
        Returns the FeatureClass name of the specified ConnectionString.
        """
        if isinstance(connection_string, list) and connection_string:
            connection_string = connection_string[0]
        if isinstance(connection_string, dict) and connection_string.get('type', '') == 'FeatureCollection':
            return 'FeatureCollection'
        if isinstance(connection_string, str) and len(connection_string) > 32:
            temp_text = connection_string.replace("\n", "").replace(" ", "")[0:32]

            if temp_text.startswith("{'type':'FeatureCollection'"):
                return 'FeatureCollection'
            if temp_text.startswith('{"type":"FeatureCollection"'):
                return 'FeatureCollection'

        if isinstance(connection_string, dict) and connection_string.get('name', ''):
            connection_string = connection_string.get('name')

        file_name, file_ext = os.path.splitext(connection_string)
        if file_name and file_ext:
            tmp_key = os.path.basename(file_name)
            tmp_pos = tmp_key.find('.')
            return tmp_key[:tmp_pos] if tmp_pos != -1 else tmp_key

        raise Exception('Unknown LayerName parsing of the ConnectionString="{}"'.format(connection_string))

    @staticmethod
    def _convert_list_args_to_dict_args(
            pipeline_args: Union[Iterable[str], Dict[str, str]]) -> Dict[str, str]:
        """
        Converts if necessary the specified type-list pipeline_args to a type-dict object.
        """
        if isinstance(pipeline_args, dict):
            return pipeline_args

        temp_args = dict()
        skip_arg = False

        for arg_index, arg in enumerate(pipeline_args):
            arg_l = arg.lower()

            if skip_arg:
                skip_arg = False
                continue
            if arg_l in ['-source', '-i', '-input', '-target', '-o', '-output'] or arg_l.startswith('--'):
                temp_args[arg] = pipeline_args[arg_index + 1]
                skip_arg = True
                continue
            else:
                temp_item = pipeline_args[arg_index]
                temp_ipos = temp_item.find('=')

                if temp_ipos != -1:
                    temp_pair = [temp_item[0:temp_ipos], temp_item[temp_ipos+1:]]
                else:
                    temp_pair = [temp_item, temp_item]

                temp_args[temp_pair[0].strip()] = temp_pair[1].strip()

        return temp_args

    @staticmethod
    def _enumerate_environment_args(objects: List[AbstractModule],
                                    pipeline_args: Dict[str, str]) -> Iterable[Tuple[str, str]]:
        """
        Enumerate the Pipeline environment args.
        """
        feature_class_found = False

        for key, value in pipeline_args.items():
            #
            if key in ['-source', '-i', '-input']:
                feature_class_found = True
                yield 'FEATURE_CLASS', PipelineManager._get_layer_name(value)
            elif key.startswith('--'):
                key = key[2:].split('.')
                if len(key) != 2:
                    continue
                object_type, attribute_name = key
                if object_type.lower() == 'pipeline':
                    yield attribute_name, value

        if not feature_class_found and objects:
            obj = objects[0]
            if isinstance(obj, AbstractReader) and hasattr(obj, 'connectionString'):
                yield 'FEATURE_CLASS', PipelineManager._get_layer_name(obj.connectionString)

        pass

    @staticmethod
    def _replace_environment_args(objects: List[AbstractModule],
                                  pipeline_args: Dict[str, str],
                                  value: Union[str, List[str]]) -> Union[str, List[str]]:
        """
        Replace the Pipeline environment args in the specified value
        """
        if isinstance(value, str):
            for key, env_value in PipelineManager._enumerate_environment_args(objects, pipeline_args):
                if os.path.isdir(env_value) or os.path.isfile(env_value):
                    import pathlib
                    p = pathlib.Path(os.path.abspath(env_value))
                    env_value = p.as_posix()

                value = value.replace('${' + key + '}', env_value)
                value = value.replace('%' + key + '%', env_value)

        elif isinstance(value, list):
            for i in range(0, len(value)):
                temp_val = PipelineManager._replace_environment_args(objects, pipeline_args, value[i])
                value[i] = temp_val

        return value

    def _parse_pipeline(self,
                        pipeline: Iterable[Dict],
                        pipeline_args: Dict[str, str] = dict()) -> List[AbstractModule]:
        """
        Parse the specified operation settings of a deserialized Pipeline.
        """
        logging.debug('Starting the parsing of Pipeline...')

        # Load the available list of modules.
        modules_root = os.path.dirname(os.path.abspath(__file__))
        manager = PipelineModuleManager()
        modules = manager.load_modules([
            os.path.join(modules_root, 'readers'),
            os.path.join(modules_root, 'filters'),
            os.path.join(modules_root, 'writers')
        ])
        if self._custom_modules_path:
            modules = manager.append_modules(self._custom_modules_path.split(','))
        #
        if self._custom_modules:
            modules.update(self._custom_modules)

        logging.debug('Loading metadata of Modules...')

        objects = PipelineManager._parse_pipeline_objects(manager, modules, pipeline, pipeline_args)
        logging.debug('Creation of Pipeline operations finished')

        return objects

    def _build_tree_pipeline(self) -> None:
        """
        Build tree of relations between Modules.
        """
        set_of_objects = {obj.stageId: obj for obj in self.objects(recursive=True)}

        def _connect_modules(a: AbstractModule, b: AbstractModule) -> None:
            """
            Connect the specified Modules.
            """
            a.ti_.outputs[b.stageId] = b
            b.ti_.inputs[a.stageId] = a

        def _build_tree_module(module_obj):
            """
            Returns the top of the specified Module.
            """
            if hasattr(module_obj, 'inputStageId'):
                obj = set_of_objects.get(module_obj.inputStageId)
                if not obj:
                    logging.warning(
                        'The StageId="{}" was not found in the tree of Objects.'
                        .format(module_obj.inputStageId))
                else:
                    _connect_modules(obj, module_obj)

            if hasattr(module_obj, 'outputStageId'):
                obj = set_of_objects.get(module_obj.outputStageId)
                if not obj:
                    logging.warning(
                        'The StageId="{}" was not found in the tree of Objects.'
                        .format(module_obj.outputStageId))
                else:
                    _connect_modules(module_obj, obj)

            if module_obj.className == 'ConnectionJoin':
                for stage_id in module_obj.stages:
                    obj = set_of_objects.get(stage_id)
                    if not obj:
                        logging.warning(
                            'The StageId="{}" was not found in the tree of Objects.'
                            .format(stage_id))
                    else:
                        _connect_modules(obj, module_obj)

                return module_obj

            a = module_obj
            children = module_obj.ti_.children

            for i in range(0, len(children)):
                b = _build_tree_module(children[i])

                if a.className != 'AbstractModule' and not isinstance(b, AbstractReader) and not b.ti_.inputs:
                    _connect_modules(a, b)

                a = b

            return a

        # Connect tree of available Modules.
        root_module = AbstractModule()
        root_module.ti_ = type('TreeMetadata', (object,), {
            'children': self._objects,
            'outputs': OrderedDict()
        })
        _build_tree_module(root_module)
        pass

    @staticmethod
    def _reassign_parameters_objects(
            manager: PipelineModuleManager,
            objects: List[AbstractModule],
            pipeline_args: Dict[str, str] = dict()) -> List[AbstractModule]:
        """
        Reassign parameters of operations.
        """
        for key, value in pipeline_args.items():
            #
            if key in ['-source', '-i', '-input']:
                data_source = value
                obj = manager.find_data_source(data_source, AbstractReader, StoreCapabilities.READ)
                if obj is None:
                    raise Exception('Input DataSource "{}" not supported!'.format(data_source))

                obj.connectionString = data_source
                objects.insert(0, obj)
                continue
            if key in ['-target', '-o', '-output']:
                data_source = value
                obj = manager.find_data_source(data_source, AbstractWriter, StoreCapabilities.CREATE)
                if obj is None:
                    raise Exception('Output DataSource "{}" not supported!'.format(data_source))

                obj.connectionString = data_source
                objects.append(obj)
                continue
            #
            elif key.startswith('--'):
                key = key[2:].split('.')
                if len(key) != 3:
                    continue
                object_type, stage_id, attribute_name = key
                if object_type.lower() != 'module':
                    continue

                logging.debug(
                    'Assigning parameter of "{}.{}" -> {}={}...'
                    .format(object_type, stage_id, attribute_name, value))

                for obj in objects:
                    if obj.stageId == stage_id:
                        if hasattr(obj, attribute_name):
                            new_value = ProcessingUtils.cast_value(value, type(getattr(obj, attribute_name)))
                            setattr(obj, attribute_name, new_value)
                        else:
                            setattr(obj, attribute_name, value)

        return objects

    @staticmethod
    def _parse_pipeline_objects(manager: PipelineModuleManager,
                                modules: Dict[str, Type],
                                pipeline: Iterable[Dict],
                                pipeline_args: Dict[str, str] = dict()) -> List[AbstractModule]:
        """
        Parse the specified operation settings of a deserialized Pipeline.
        """
        objects: List[AbstractModule] = list()
        temp_args = {k: v for k, v in pipeline_args.items() if k.startswith('--') or not k.startswith('-')}

        # Create the operation object collection of the pipeline.
        for settings in pipeline:
            type_name = settings['type']

            type_parts = type_name.split('.')
            class_name = type_parts[len(type_parts)-1]
            type_def = modules.get(type_name.lower()) or modules.get(class_name.lower())

            # ... does this object provide the Path where the Python file is located?
            if type_def is None and settings.get('moduleLocation'):
                script_file = settings.get('moduleLocation')

                if not os.path.exists(script_file):
                    script_file = os.path.join(os.path.dirname(__file__), script_file)

                type_def = ModuleManager.import_type_from_file(script_file, class_name, inspect.isclass)

            if type_def is None:
                raise Exception('The Module type "{}" is not supported!'.format(settings['type']))

            obj = type_def()
            obj.ti_ = type('TreeMetadata', (object,), {
                'parent': None, 'children': [], 'inputs': OrderedDict(), 'outputs': OrderedDict()
            })

            for key, value in settings.items():
                if key not in ['type', 'pipeline']:
                    setattr(obj, key, value)

            if settings.get('pipeline'):
                children_pipeline = settings.get('pipeline')
                children = \
                    PipelineManager._parse_pipeline_objects(manager, modules, children_pipeline, temp_args)

                for child in children:
                    child.ti_.parent = obj

                obj.ti_.children = children

            objects.append(obj)

        # Reassign parameters of operations.
        objects = \
            PipelineManager._reassign_parameters_objects(manager, objects, pipeline_args)

        # Fix some possible bad settings.
        for obj in objects:
            if isinstance(obj, (AbstractReader, AbstractWriter)) and hasattr(obj, 'connectionString'):
                obj.connectionString = \
                    PipelineManager._replace_environment_args(objects, pipeline_args, obj.connectionString)

        return objects

    def load_from_file(self,
                       file_name: str,
                       pipeline_args: Union[Iterable[str], Dict[str, str]] = dict()) -> List[AbstractModule]:
        """
        Load the pipeline from the specified JSON filename.
        """
        self._pipeline_dir = os.path.dirname(file_name)
        self._objects = list()
        pipeline_args = PipelineManager._convert_list_args_to_dict_args(pipeline_args)

        with open(file_name, 'r') as file:
            # remove JSON comments.
            standard_json = JsonComments.remove_comments(file, keep_ends=True)
            # Reassign environment pipeline content.
            standard_json = PipelineManager._replace_environment_args([], pipeline_args, standard_json)

            # Reassign pipeline content from current file.
            import pathlib
            p = pathlib.Path(os.path.abspath(file_name))
            model_path = p.as_posix()
            model_dirn = os.path.dirname(model_path)
            model_file = os.path.basename(model_path)

            file_args = {
                "PIPELINE_FOLDER": model_dirn,
                "PIPELINE_FILE": model_file,
                "PIPELINE_PATH": model_path
            }
            for k, v in pipeline_args.items():
                if not k.startswith('-'):
                    standard_json = standard_json.replace('$'+k, v)
            for k, v in file_args.items():
                standard_json = standard_json.replace('$'+k, v)

            pipeline = json.loads(standard_json)['pipeline']
            self._objects = self._parse_pipeline(pipeline, pipeline_args)
            self._build_tree_pipeline()

        return self._objects

    def load_from_json(self,
                       pipeline: Iterable[Dict],
                       pipeline_args: Union[Iterable[str], Dict[str, str]] = dict()) -> List[AbstractModule]:
        """
        Load the pipeline from the specified JSON pipeline.
        """
        self._pipeline_dir = None
        self._objects = list()
        pipeline_args = PipelineManager._convert_list_args_to_dict_args(pipeline_args)

        self._objects = self._parse_pipeline(pipeline, pipeline_args)
        self._build_tree_pipeline()
        return self._objects

    def get_schema(self, processing_args, stageId: str) -> bool:
        """
        Get the Schema of a Stage in the specified pipeline of Geospatial data.
        """
        temp_list = [obj for obj in self.objects(recursive=True) if obj.stageId == stageId]
        if not temp_list:
            raise Exception('Stage "{}" not found in current Pipeline'.format(stageId))

        class DummyWriter(AbstractWriter):
            """
            DummyWriter to capture SchemaDef.
            """
            def __init__(self):
                AbstractWriter.__init__(self)
                self.schema_def = None

            def test_capability(self, connection_string: str, capability: StoreCapabilities) -> bool:
                """
                Returns always True for this DummyWriter.
                """
                return True

            def starting_run(self, schema_def, pipeline, processing_args):
                """
                Capture SchemaDef.
                """
                self.schema_def = schema_def.clone()
                return schema_def

            def run(self, data_store, processing_args):
                """
                Doing nothing.
                """
                return iter([])

        current_list = self._objects
        try:
            new_list = list()

            dummy_writer = DummyWriter()
            dummy_writer.ti_ = type('TreeMetadata', (object,), {
                'parent': None, 'children': [], 'inputs': OrderedDict(), 'outputs': OrderedDict()
            })

            def my_clone_of_object_function(obj) -> Iterable:
                #
                if obj.stageId == stageId:
                    #
                    if isinstance(obj, AbstractWriter):
                        dummy_writer.ti_.inputs = obj.ti_.inputs.copy()
                        yield dummy_writer
                    else:
                        new_obj = deepcopy(obj)
                        new_obj.ti_.outputs.clear()
                        new_obj.ti_.outputs[dummy_writer.stageId] = dummy_writer
                        yield new_obj
                        dummy_writer.ti_.inputs.clear()
                        dummy_writer.ti_.inputs[new_obj.stageId] = new_obj
                        yield dummy_writer

                        new_obj.ti_.children.clear()
                        for child in obj.ti_.children:
                            for c in my_clone_of_object_function(child):
                                new_obj.ti_.children[c.stageId] = c

                    return

                if isinstance(obj, AbstractWriter):
                    return

                yield obj

            for item in self._objects:
                new_list.extend(my_clone_of_object_function(item))

            self._objects = new_list
            self.run(processing_args)

            return dummy_writer.schema_def
        finally:
            self._objects = current_list

    def run(self, processing_args, callback: Callable = None, callback_args: Any = None) -> bool:
        """
        Execute the pipeline of Geospatial data.
        """
        if len(self._objects) == 0:
            logging.warning('There is none Node, the Pipeline does nothing!')
            return False

        set_of_objects = [obj for obj in self.objects(recursive=True)]

        if self._pipeline_dir:
            os.environ['PIPELINE_FOLDER'] = self._pipeline_dir
            sys.path.append(self._pipeline_dir)

        try:
            readers = [
                obj for obj in set_of_objects if isinstance(obj, AbstractReader)
            ]
            writers = [
                obj for obj in set_of_objects if isinstance(obj, AbstractWriter) and not obj.ti_.outputs
            ]

            if not writers:
                logging.warning('There is none Output node, the Pipeline does nothing!')
                return False

            # Redefine 'connectionString' of Readers, when defining embebed fileData.
            for reader in readers:
                if hasattr(reader, 'connectionString'):
                    tempString = reader.connectionString

                    if isinstance(tempString, dict) and tempString.get('name') and tempString.get('fileData'):
                        file_data = tempString.get('fileData')
                        temp_ipos = file_data.find(';base64,')
                        file_data = file_data[temp_ipos+len(';base64,'):]
                        file_name = tempString.get('name')
                        temp_file = os.path.join(
                            processing_args.temp_data_path(), '{}_{}'.format(uuid.uuid1(), file_name))

                        with open(temp_file, mode='wb') as fp:
                            stream_fp = base64.b64decode(file_data)
                            fp.write(stream_fp)

                        reader.connectionString = temp_file

            # Run workflow, like one IEnumerable stream!
            for writer in writers:
                self._invoke_starting_run(writer, [None, None], processing_args)

                for feature in writer:
                    if callback:
                        callback(self, processing_args, writer, feature, callback_args)

                    feature = None

                self._invoke_finished_run(writer, processing_args)

            return True
        finally:
            for obj in set_of_objects:
                if hasattr(obj, 'pipeline_args'):
                    obj.finished_run(self, processing_args)
                    delattr(obj, 'pipeline_args')
                if hasattr(obj, 'clean') and callable(getattr(obj, 'clean')):
                    obj.clean()

            if self._pipeline_dir:
                os.environ['PIPELINE_FOLDER'] = ''
                sys.path.remove(self._pipeline_dir)

        return False

    def _invoke_starting_run(
            self,
            module_obj: AbstractModule,
            function_args: Any, processing_args) -> Tuple[SchemaDef, AbstractModule]:
        """
        Prepare of task metadata of each Pipeline Operation.
        """
        input_function_args = function_args
        input_schemas = []

        # Calling before to the tree of inputs.
        for obj in module_obj.ti_.inputs.values():
            function_args = self._invoke_starting_run(obj, input_function_args, processing_args)
            input_schemas.append(function_args[0])

        if hasattr(module_obj, 'pipeline_args'):
            return [module_obj.pipeline_args.schema_def, module_obj.pipeline_args.data_source]
        if len(input_schemas) > 1:
            function_args[0] = SchemaDef.merge_all(input_schemas)
        if module_obj.className == 'ConnectionJoin':
            function_args[1] = None

        # Assign metadata for using when the Pipeline runs.
        schema_def = function_args[0]
        parent_obj = function_args[1]
        schema_def = module_obj.starting_run(schema_def, self, processing_args)

        setattr(module_obj, 'pipeline_args', type('PipelineArgs', (object,), {
            'config': self.config,
            'pipeline': self,
            'data_source': parent_obj,
            'schema_def': schema_def,
            'processing_args': processing_args
        })())
        return [schema_def, module_obj]

    def _invoke_finished_run(self, module_obj: AbstractModule, processing_args) -> AbstractModule:
        """
        Finalization of task for each Pipeline operation.
        """
        if hasattr(module_obj, 'pipeline_args'):
            module_obj.finished_run(self, processing_args)
            delattr(module_obj, 'pipeline_args')

        for obj in module_obj.ti_.inputs.values():
            self._invoke_finished_run(obj, processing_args)

        return module_obj



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from datetime import date, datetime
from typing import Any, List


class DataType(object):
    """
    List of available Data Types of Attributes or Columns (Identifiers and types
    match the OGR/Fiona type specs).
    """
    """
    + OGR OFT enum: [
        ogr.OFTInteger,
        ogr.OFTIntegerList,
        ogr.OFTReal,
        ogr.OFTRealList,
        ogr.OFTString,
        ogr.OFTStringList,
        ogr.OFTWideString,
        ogr.OFTWideStringList,
        ogr.OFTBinary,
        ogr.OFTDate,
        ogr.OFTTime,
        ogr.OFTDateTime,
        ogr.OFTInteger64,
        ogr.OFTInteger64List
        ]
    + fiona.ogrext.FIELD_TYPES: [
        'int32', None, 'float', None, 'str', None, None, None, 'bytes',
        'date', 'time', 'datetime', 'int64', None
        ]
    + fiona.ogrext.FIELD_TYPES_MAP: {
        'int32': <class 'int'>,
        'float': <class 'float'>,
        'str': <class 'str'>,
        'bytes': <class 'bytes'>,
        'date': <class 'fiona.rfc3339.FionaDateType'>,
        'time': <class 'fiona.rfc3339.FionaTimeType'>,
        'datetime': <class 'fiona.rfc3339.FionaDateTimeType'>,
        'int64': <class 'int'>,
        'int': <class 'int'>
        }
    """
    Integer = 0
    Float = 2
    String = 4
    Binary = 8
    Date = 9
    Time = 10
    DateTime = 11
    Integer64 = 12

    @staticmethod
    def to_data_type(value: Any) -> "DataType":
        """
        Returns the DataType of the specified value.
        """
        if isinstance(value, int):
            return DataType.Integer
        if isinstance(value, float):
            return DataType.Float
        if isinstance(value, str):
            return DataType.String
        if isinstance(value, bytes):
            return DataType.Binary
        if isinstance(value, date):
            return DataType.Date
        if isinstance(value, datetime):
            return DataType.DateTime
        if isinstance(value, bool):
            return DataType.Integer

        return DataType.String


class GeometryType(object):
    """
    List of available Geometry Types (Identifiers and types match the OGR/Fiona WKB type specs).
    """
    """
    + OGR GT enum: [
        ogr.wkbUnknown,
        ogr.wkbPoint, ogr.wkbPointM (wkb+2000), ogr.wkbPointZM (wkb+3000),
        ogr.wkbLineString, ...
        ogr.wkbPolygon,
        ogr.wkbMultiPoint,
        ogr.wkbMultiLineString,
        ogr.wkbMultiPolygon,
        ogr.wkbGeometryCollection
        ]
    """
    Unknown = 0
    Point = 1
    LineString = 2
    Polygon = 3
    MultiPoint = 4
    MultiLineString = 5
    MultiPolygon = 6
    GeometryCollection = 7


class FieldDef:
    """
    Provides metadata information of a Geospatial Attribute or Column.
    """
    def __init__(self,
                 name: str,
                 data_type: DataType,
                 precision: int = 0,
                 width: int = 0,
                 is_nullable: bool = True,
                 default_value: Any = None,
                 **kwargs):
        self.name = name
        self.type = data_type
        self.precision = precision
        self.width = width
        self.nullable = is_nullable
        self.defaultValue = default_value
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])

    def __str__(self) -> str:
        """
        Returns the String representation of this Object.
        """
        return 'Name = {}, DataType = {}, Precision = {}, Width = {}, Nullable = {}, Default = {}' \
               .format(self.name,
                       self.type,
                       self.precision,
                       self.width,
                       self.nullable,
                       self.defaultValue)

    def copy(self, field_def: "FieldDef") -> None:
        """
        Copy the metadata information of specified input FieldDef.
        """
        for k, v in field_def.__dict__.items():
            setattr(self, k, v)

    def clone(self) -> "FieldDef":
        """
        Clone this FieldDef object.
        """
        new_obj = FieldDef(name=self.name, data_type=self.type)
        new_obj.copy(self)
        return new_obj


class SchemaDef:
    """
    Provides metadata information of a Geospatial Dataset.
    """
    def __init__(self, **kwargs):
        self.name = ''
        self.type = ''
        self.srid = 0
        self.crs = None
        self.geometryType = None
        self.envelope = None
        self.fields = list()
        for arg in kwargs:
            setattr(self, arg, kwargs[arg])

    def copy(self, schema_def: "SchemaDef") -> None:
        """
        Copy the metadata information of specified input SchemaDef.
        """
        for k, v in schema_def.__dict__.items():
            setattr(self, k, v)

        self.fields = [fd.clone() for fd in schema_def.fields]

    def clone(self) -> "SchemaDef":
        """
        Clone this SchemaDef object.
        """
        new_obj = SchemaDef()
        new_obj.copy(self)
        return new_obj

    def merge(self, other: "SchemaDef") -> "SchemaDef":
        """
        Merge metadata of this Object and specified input SchemaDef.
        """
        if not other:
            return self

        a, b = self, other

        geometry_type_a = a.geometryType
        geometry_type_b = b.geometryType
        if geometry_type_a != geometry_type_b:
            raise Exception('Merging SchemaDefs of different GeometryType is not supported.')
        srid_a = a.srid
        srid_b = b.srid
        if srid_a != srid_b:
            raise Exception('Merging SchemaDefs of different CRS is not supported.')

        from geodataflow.core.common import CaseInsensitiveDict
        fields1 = b.fields
        fd_hash = CaseInsensitiveDict({field.name: field for field in a.fields})
        fields1 = [f.clone() for f in fields1 if fd_hash.get(f.name, None) is None]
        if not fields1:
            return a
        schema_def = a.clone()
        schema_def.fields = schema_def.fields + fields1
        return schema_def

    @staticmethod
    def merge_all(schemas: List["SchemaDef"]) -> "SchemaDef":
        """
        Merge metadata of specified stream of input SchemaDefs.
        """
        if not schemas:
            return None

        a = schemas[0]
        for b in schemas[1:]:
            a = a.merge(b)

        return a



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import logging
from typing import Dict

from geodataflow.core.capabilities import StoreCapabilities
from geodataflow.pipeline.basictypes import AbstractReader


class FeatureReader(AbstractReader):
    """
    Reads Features with Geometries from a Geospatial DataSource using OGR providers.
    """
    def __init__(self):
        AbstractReader.__init__(self)
        self._featureStore = None
        self.connectionString = ''
        self.where = ''
        self.spatialFilter = ''
        self.countLimit = -1

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Reads Features with Geometries from a Geospatial DataSource using OGR providers.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Input'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'connectionString': {
                'description': 'Connection string of the Feature Store.',
                'dataType': ['file', 'url', 'geojson'],
                'extensions': ['.geojson', '.gpkg', '.shp.zip']
            },
            'where': {
                'description': 'Attribute query string when fetching features (Optional).',
                'dataType': 'filter'
            },
            'spatialFilter': {
                'description': 'Geometry to be used as spatial filter when fetching features (Optional).',
                'dataType': 'geometry'
            },
            'countLimit': {
                'description': 'Maximum number of Features to fetch (Optional).',
                'dataType': 'int'
            }
        }

    def test_capability(self, connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns of this Module supports the specified ConnectionString and named StoreCapability.
        """
        from geodataflow.geoext.commonutils import GdalUtils
        return GdalUtils.test_ogr_capability(connection_string, capability)

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.geoext.featurestore import OgrFeatureStore

        # Open the FeatureStore.
        self._featureStore = OgrFeatureStore()
        schema_def = self._featureStore.open(self.connectionString, 'r', self.spatialFilter, self.where)
        return schema_def

    def run(self, none_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        from geodataflow.geoext.commonutils import DataUtils
        from geodataflow.pipeline.progress import ProgressProcessingStore

        # Show UI percentage progress?
        ui_progress = ProgressProcessingStore(processing_args)

        connection_string = list(DataUtils.enumerate_single_connection_string(self.connectionString))[0]
        is_geojson = DataUtils.represents_json_feature_collection(connection_string)
        if is_geojson:
            connection_string = 'GeoJSON FeatureCollection'

        logging.info('Reading Features from "{}"...'.format(connection_string))

        feature_layers = self._featureStore.layers()
        feature_count = 0
        count_limit = self.countLimit if self.countLimit is not None else -1

        # Enumerate Features...
        for feature in self._featureStore.features():
            #
            if ui_progress.enabled:
                if feature_count == 0:
                    total_count = 0
                    for feature_layer in feature_layers:
                        total_count += feature_layer.layer().GetFeatureCount()

                    ui_progress.initialize(total_count)
                else:
                    ui_progress.update()

            if count_limit > feature_count:
                logging.warning(
                    'Early processing break! The "countLimit" flag was specified (Value={}).'
                    .format(self.countLimit))

                break

            # DEBUG: if feature_count == 100: break
            feature_count += 1
            yield feature

        if ui_progress.enabled and ui_progress.initialized():
            logging.info('')

        logging.info('{:,} Features read from "{}".'.format(feature_count, connection_string))
        pass

    def finished_run(self, pipeline, processing_args):
        """
        Finishing a Workflow on Geospatial data.
        """
        self._featureStore = None
        return True



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import logging
import datetime
from typing import Dict, Iterable

from geodataflow.core.common import CaseInsensitiveDict
from geodataflow.pipeline.basictypes import AbstractFilter


class EOProductCatalog(AbstractFilter):
    """
    The Filter extracts Metadata from EO/STAC Collections via spatial & alphanumeric filters.
    """
    def __init__(self):
        AbstractFilter.__init__(self)
        self.driver = 'STAC'
        self.provider = 'https://earth-search.aws.element84.com/v0/search'
        self.product = 'sentinel-s2-l2a-cogs'
        self.startDate = None
        self.endDate = None
        self.closestToDate = None
        self.windowDate = None
        self.filter = None
        self.preserveInputCrs = True

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'Catalog'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Extracts Metadata from EO/STAC Collections via spatial & alphanumeric filters.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'EO STAC Imagery'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'driver': {
                'description': 'Driver class name that implements EO Providers.',
                'dataType': 'string',
                'default': 'STAC',
                'options': ['STAC', 'EODAG'],
                'labels': ['STAC', 'EODAG']
            },
            'provider': {
                'description': 'Provider name or API Endpoint that provides info about EO Products.',
                'dataType': 'string',
                'default': 'https://earth-search.aws.element84.com/v0/search'
            },
            'product': {
                'description': 'EO Product type or Collection from which to fetch data.',
                'dataType': 'string',
                'default': 'sentinel-s2-l2a-cogs'
            },
            'startDate': {
                'description': 'Start date of EO Products to fetch (Optional). "$TODAY()" is supported.',
                'dataType': 'date'
            },
            'endDate': {
                'description': 'End date of EO Products to fetch (Optional). "$TODAY()" is supported.',
                'dataType': 'date'
            },
            'closestToDate': {
                'description':
                    'Select only those EO Products which Date is the closest to the specified (Optional).',
                'dataType': 'date'
            },
            'windowDate': {
                'description':
                    'Days around "closestToDate" when "startDate" and "endDate" are not specified.',
                'dataType': 'int',
                'default': 5
            },
            'filter': {
                'description': 'Attribute filter string of EO Products to fetch (Optional).',
                'dataType': 'filter'
            },
            'preserveInputCrs': {
                'description': 'Preserve input CRS, otherwise Geometries are transformed to "EPSG:4326".',
                'dataType': 'bool',
                'default': True
            }
        }

    def fetch_products(self, geometry, input_crs, app_config: Dict, limit: int = 1000) -> Iterable:
        """
        Ferch EO Products using current settings.
        """
        today_date = datetime.date.today()

        if isinstance(self.startDate, str):
            start_date = today_date \
                if self.startDate.upper() == '$TODAY()' \
                else datetime.datetime.strptime(self.startDate, '%Y-%m-%d')

        elif self.startDate is None and self.closestToDate:
            temps_time = datetime.datetime.strptime(self.closestToDate, '%Y-%m-%d')
            start_date = self.closestToDate \
                if not self.windowDate \
                else (temps_time - datetime.timedelta(days=self.windowDate)).strftime('%Y-%m-%d')

            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
        else:
            start_date = self.startDate if self.startDate else today_date

        if isinstance(self.endDate, str):
            final_date = today_date \
                if self.endDate.upper() == '$TODAY()' \
                else datetime.datetime.strptime(self.endDate, '%Y-%m-%d')

        elif self.startDate is None and self.closestToDate:
            temps_time = datetime.datetime.strptime(self.closestToDate, '%Y-%m-%d')
            final_date = self.closestToDate \
                if not self.windowDate \
                else (temps_time + datetime.timedelta(days=self.windowDate)).strftime('%Y-%m-%d')

            final_date = datetime.datetime.strptime(final_date, '%Y-%m-%d')
        else:
            final_date = self.endDate if self.endDate else today_date

        product_filter = dict()
        if self.filter:
            for item in self.filter.replace('==', '=').split(';'):
                pair = item.split('=')
                product_filter[pair[0].strip()] = pair[1].strip()

        # Search available EO products according the current AOI-TimeRange criteria.
        from geodataflow.eogeo.productcatalog import ProductCatalog
        eo_catalog = ProductCatalog()

        # Transform Geometries to EPSG:4326 for searching EO Products?
        from geodataflow.geoext.commonutils import GeometryUtils
        transform_fn = GeometryUtils.create_transform_function(input_crs, 4326)
        geometry = transform_fn(geometry)

        params = dict(
            driver=self.driver,
            provider=self.provider,
            config=app_config,
            product_type=self.product,
            area=geometry,
            date=(start_date, final_date),
            limit=limit
        )
        if len(product_filter) > 0:
            params.update(product_filter)

        # Search EO Products!
        for product in eo_catalog.search(**params):
            yield product

        pass

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.core.schemadef import DataType, GeometryType, FieldDef
        from geodataflow.geoext.commonutils import GeometryUtils

        envelope = schema_def.envelope if schema_def.envelope else [0, -90, 360, 90]
        geometry = GeometryUtils.create_geometry_from_bbox(envelope[0], envelope[1], envelope[2], envelope[3])

        # Fetch schema from first EO Product.
        app_config = pipeline.config
        field_dict = {f.name: f for f in schema_def.fields}
        new_fields = [
            FieldDef(name='gdf:product_date', data_type=DataType.String)
        ]
        for product in self.fetch_products(geometry, schema_def.crs, app_config, limit=1):
            properties = product.properties

            for name, value in properties.items():
                if not field_dict.get(name):
                    new_fields.append(FieldDef(name=name, data_type=DataType.to_data_type(value)))

        schema_def = schema_def.clone()
        schema_def.geometryType = GeometryType.Polygon
        schema_def.input_srid = schema_def.srid
        schema_def.input_crs = schema_def.crs
        schema_def.fields += new_fields

        # Redefine CRS when distinct of EPSG:4326?
        if not self.preserveInputCrs and schema_def.srid != 4326 and schema_def.envelope:
            from pyproj import CRS
            from geodataflow.geoext.commonutils import GeometryUtils

            schema_def.srid = 4326
            schema_def.crs = CRS.from_epsg(4326)

            transform_fn = GeometryUtils.create_transform_function(schema_def.input_crs, schema_def.crs)
            geometry = GeometryUtils.create_geometry_from_bbox(*schema_def.envelope)
            geometry = transform_fn(geometry)
            schema_def.envelope = list(geometry.bounds)

        return schema_def

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        schema_def = self.pipeline_args.schema_def
        app_config = self.pipeline_args.config

        report_info = {
            'driver': self.driver,
            'provider': self.provider,
            'productType': self.product,
            'bounds': '',
            'startDate': self.startDate,
            'endDate': self.endDate,
            'closestToDate': self.closestToDate if self.closestToDate else '',
            'filter': self.filter if self.filter else '',
            'availableDates': [],
            'selectedDate': ''
        }

        # Transform Geometries to EPSG:4326 for searching EO Products?
        from geodataflow.geoext.commonutils import GeometryUtils
        inv_transform_fn = \
            GeometryUtils.create_transform_function(4326, schema_def.input_crs) \
            if self.preserveInputCrs else None

        for feature in feature_store:
            geometry = feature.geometry
            report_info['bounds'] = 'BBOX{}'.format(geometry.bounds)

            # Search EO Products!
            eo_products = [
                self.normalize_product(product, feature)
                for product in self.fetch_products(geometry, schema_def.input_crs, app_config, limit=1000)
            ]
            logging.info("Available {} EO Products for type '{}'.".format(len(eo_products), self.product))

            available_dates = set([product.properties.get('gdf:product_date') for product in eo_products])
            report_info['availableDates'] = list(available_dates)

            # Return results.
            for product in self.pass_products(eo_products):
                report_info['selectedDate'] = product.properties.get('gdf:product_date')
                if inv_transform_fn:
                    product.geometry = inv_transform_fn(product.geometry)

                yield product
            #

        report_context = \
            processing_args.reportContext if hasattr(processing_args, 'reportContext') else None
        if report_context:
            setattr(report_context, self.className, report_info)

        pass

    def pass_products(self, products: Iterable[object]) -> Iterable[object]:
        """
        Returns only those EO Products that match custom criteria (e.g. "closestToDate").
        """
        if self.closestToDate:
            closest_time = datetime.datetime.strptime(self.closestToDate, '%Y-%m-%d')
            temp_dict = dict()
            temp_list = list()

            for product in products:
                product_date = product.properties.get('gdf:product_date')
                product_time = datetime.datetime.strptime(product_date, '%Y-%m-%d')
                product_diff = abs(product_time - closest_time).days
                temp_list.append((product_date, product_time, product_diff))
                curr_list = temp_dict.get(product_date, [])
                curr_list.append(product)
                temp_dict[product_date] = curr_list

            for best_item in sorted(temp_list, key=lambda tup: tup[2], reverse=False):
                best_date = best_item[0]
                best_list = temp_dict[best_date]
                return best_list

        return products

    def normalize_product(self, product: object, feature: object) -> object:
        """
        Normalize properties of specified EOProduct.
        """
        case_props = CaseInsensitiveDict(product.properties)
        attributes = feature.properties.copy()

        product_date = \
            case_props.get('startTimeFromAscendingNode') or \
            case_props.get('beginPosition') or \
            case_props.get('ingestionDate') or \
            case_props.get('publicationDate') or \
            case_props.get('datetime')

        attributes['gdf:product_date'] = product_date[0:10]

        for name, value in product.properties.items():
            attributes[name] = value if not isinstance(value, list) else ','.join([str(v) for v in value])

        product.areaOfInterest = feature
        product.properties = attributes

        return product



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from typing import Dict
from geodataflow.pipeline.basictypes import AbstractFilter


class GeometryTransform(AbstractFilter):
    """
    The Filter transforms input Geometries between two Spatial Reference Systems (CRS).
    """
    def __init__(self):
        AbstractFilter.__init__(self)
        self.sourceCrs = None
        self.targetCrs = None

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'Transform'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Transforms input Geometries or Rasters between two Spatial Reference Systems (CRS).'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Geometry'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'sourceCrs': {
                'description':
                    'Source Spatial Reference System (CRS), SRID, WKT, PROJ formats are supported. ' +
                    'It uses input CRS when this param is not specified.',
                'dataType': 'crs',
                'placeHolder': 'EPSG:XXXX or SRID...'
            },
            'targetCrs': {
                'description':
                    'Output Spatial Reference System (CRS), SRID, WKT, PROJ formats are supported.',
                'dataType': 'crs',
                'placeHolder': 'EPSG:XXXX or SRID...'
            }
        }

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.geoext.commonutils import GeometryUtils

        if self.targetCrs:
            source_crs = GeometryUtils.get_spatial_crs(self.sourceCrs if self.sourceCrs else schema_def.crs)
            target_crs = GeometryUtils.get_spatial_crs(self.targetCrs)

            schema_def = schema_def.clone()
            schema_def.input_srid = source_crs.to_epsg()
            schema_def.input_crs = source_crs
            schema_def.srid = target_crs.to_epsg()
            schema_def.crs = target_crs

            if schema_def.envelope:
                transform_fn = GeometryUtils.create_transform_function(source_crs, target_crs)
                geometry = GeometryUtils.create_geometry_from_bbox(*schema_def.envelope)
                geometry = transform_fn(geometry)
                schema_def.envelope = list(geometry.bounds)
        else:
            schema_def = schema_def.clone()
            schema_def.input_srid = schema_def.crs.to_epsg()
            schema_def.input_crs = schema_def.crs

        return schema_def

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        from geodataflow.geoext.commonutils import GeometryUtils
        from geodataflow.geoext.dataset import GdalDataset

        schema_def = self.pipeline_args.schema_def
        source_crs = GeometryUtils.get_spatial_crs(
            self.sourceCrs if self.sourceCrs else schema_def.input_srid
        )
        target_crs = GeometryUtils.get_spatial_crs(
            self.targetCrs
        )

        if source_crs and target_crs and source_crs.to_epsg() != target_crs.to_epsg():
            transform_fn = GeometryUtils.create_transform_function(source_crs, target_crs)

            for feature in feature_store:
                if isinstance(feature, GdalDataset):
                    dataset = feature.warp(output_crs=target_crs, output_geom=None)
                    yield dataset
                else:
                    feature.geometry = transform_fn(feature.geometry)
                    yield feature
            #
        else:
            for feature in feature_store:
                yield feature

        pass



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from geodataflow.pipeline.basictypes import AbstractFilter


class FeatureCache(AbstractFilter):
    """
    The Filter caches data of inputs to speedup the management of repetitive invocations of Modules.
    """
    def __init__(self):
        AbstractFilter.__init__(self)
        self._dataCache = None

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'Cache'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Caches data of inputs to speedup the management of repetitive invocations of Modules.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Feature'

    def run(self, data_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        if self._dataCache:
            for feature in self._dataCache:
                yield feature
        else:
            data_cache = []

            for feature in data_store:
                data_cache.append(feature)
                yield feature

            self._dataCache = data_cache

        pass

    def finished_run(self, pipeline, processing_args):
        """
        Finishing a Workflow on Geospatial data.
        """
        return True

    def clean(self):
        """
        Clean cache of inputs.
        """
        if self._dataCache:
            self._dataCache.clear()
            self._dataCache = None

        return True



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import re
from typing import Iterable


class JsonComments:
    """
    A wrapper to JSON parsers allowing comments, multiline strings and trailing commas.
    See:
      https://bitbucket.org/Dando_Real_ITA/json-comment/src/default/jsoncomment/comments.py
    """
    COMMENT_PREFIX = ("#", ";", "//")
    MULTILINE_START = "/*"
    MULTILINE_END = "*/"
    LONG_STRING = '"""'
    TEMPLATE_RE = re.compile(r"\{\{(.*?)\}\}")

    @staticmethod
    def remove_comments(lines: Iterable[str], keep_ends: bool = True) -> str:
        """
        Returns the standard JSON version (Skipping comments) of the specified stream of JSON lines.
        """
        standard_json = ''
        is_multiline = False

        for line in lines:
            # 0 if there is no trailing space, 1 otherwise.
            keep_trail_space = int(line.endswith(' '))

            # Count leading spaces.
            leading_spaces = len(line) - len(line.lstrip())

            # Remove all whitespace on both sides.
            line = line.strip()

            # Skip blank lines or single line comments.
            if len(line) == 0 or line.startswith(JsonComments.COMMENT_PREFIX):
                continue

            # Mark the start of a multiline comment:
            # Not skipping, to identify single line comments using multiline comment tokens, like
            # /***** Comment *****/
            if line.startswith(JsonComments.MULTILINE_START):
                is_multiline = True

            # Skip a line of multiline comments
            if is_multiline:
                # Mark the end of a multiline comment
                if line.endswith(JsonComments.MULTILINE_END):
                    is_multiline = False

                continue

            # Replace the multiline data token to the JSON valid one.
            if JsonComments.LONG_STRING in line:
                line = line.replace(JsonComments.LONG_STRING, '"')

            standard_json += (leading_spaces * ' ') + line + ' ' * keep_trail_space
            if keep_ends:
                standard_json += '\n'

        # Removing non-standard trailing commas.
        standard_json = standard_json.replace(',]', ']')
        standard_json = standard_json.replace(',}', '}')

        return standard_json



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

from typing import Dict
from geodataflow.core.processingargs import ProcessingUtils
from geodataflow.geoext.relationships import SpatialRelationships
from geodataflow.pipeline.basictypes import AbstractFilter
from geodataflow.pipeline.filters.InputParam import InputParam


class SpatialRelation(AbstractFilter):
    """
    The Filter returns input Features that match a Spatial Relationship with one or more other Geometries.
    """
    def __init__(self):
        AbstractFilter.__init__(self)
        self.relationship = SpatialRelationships.Intersects
        self.otherGeometries = ''

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Returns input Features that match a Spatial Relationship with one or more other Geometries.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Geometry'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        temp_dict = ProcessingUtils.object_as_dict(SpatialRelationships)

        return {
            'relationship': {
                'description': "Spatial Relationship to validate, 'Intersects' by default.",
                'dataType': 'int',
                'default': SpatialRelationships.Intersects,
                'options': [v for _, v in temp_dict.items()],
                'labels': [k for k, _ in temp_dict.items()]
            },
            'otherGeometries': {
                'description': 'Collection of Geometries with which input Features should validate a Spatial Relationship.',
                'dataType': 'input'
            }
        }

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        relationship = ProcessingUtils.cast_enum(self.relationship, SpatialRelationships)
        others = [
            obj.geometry for obj in InputParam.enumerate_inputs(self.otherGeometries, self.pipeline_args)
        ]

        # Set names of implemented Spatial Relationsips between geometries.
        rel_names = {
            SpatialRelationships.Equals: 'equals',
            SpatialRelationships.Disjoint: 'disjoint',
            SpatialRelationships.Intersects: 'intersects',
            SpatialRelationships.Touches: 'touches',
            SpatialRelationships.Crosses: 'crosses',
            SpatialRelationships.Within: 'within',
            SpatialRelationships.Contains: 'contains',
            SpatialRelationships.Overlaps: 'overlaps'
        }
        rel_name = rel_names.get(relationship)
        if not rel_name:
            raise Exception('The Relationship "{}" is not supported!'.format(relationship))

        # Perform spatial relation between Geometries.
        for feature in feature_store:
            geometry = feature.geometry

            for other in others:
                if bool(geometry.impl[rel_name](geometry, other)):
                    yield feature
                    break

        pass



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import logging
import importlib
import inspect
from typing import Dict, Iterable, Type


class ModuleManager:
    """
    Module manager of dynamic modules deployed in the 'Factories' folder application.
    """
    def __init__(self, abstract_types=()):
        self.abstract_types = abstract_types
        self.modules: Dict[str, Type] = dict()

    def load_modules(self, modules_folders: Iterable[str] = []) -> Dict[str, Type]:
        """
        Load the modules deployed in the specified 'factories' folder application.
        """
        self.modules.clear()
        return self.append_modules(modules_folders)

    def append_modules(self, modules_folders: Iterable[str] = []) -> Dict[str, Type]:
        """
        Adds the modules deployed in the specified 'factories' folder application.
        """
        base_folder = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

        for modules_folder in modules_folders:
            modules_folder = os.path.abspath(modules_folder)

            if not os.path.exists(modules_folder):
                logging.error('The specified modules folder "{}" does not exist!'.format(modules_folder))
                continue

            modules_prefix = os.path.commonprefix([base_folder, modules_folder])
            modules_prefix = modules_folder[len(modules_prefix):]
            modules_prefix = modules_prefix.replace('\\', '.')
            modules_prefix = modules_prefix.replace('/', '.')
            if modules_prefix.startswith('.'):
                modules_prefix = os.path.basename(base_folder) + modules_prefix

            # Recursively reading of the factory sub-folder.
            module_file_list = os.listdir(modules_folder)
            module_list = []
            for module_file in module_file_list:
                file_name = os.path.join(modules_folder, module_file)

                if file_name.endswith('__init__.py') or file_name.endswith('__pycache__'):
                    continue
                if file_name.endswith('.py'):
                    module_name = module_file.split('.')[0]
                    module_list.append(module_name)

            # Load the modules of a sub-folder.
            for module_name in module_list:
                try:
                    module_def = importlib.import_module(modules_prefix + '.' + module_name)

                    for type_name, type_def in inspect.getmembers(module_def, inspect.isclass):
                        if type_def in self.abstract_types or not issubclass(type_def, self.abstract_types):
                            continue

                        if hasattr(type_def, '__abstractmethods__') and type_def.__abstractmethods__:
                            continue

                        type_name = type_name.lower()
                        self.modules[type_name] = type_def

                except Exception as e:
                    logging.error('Fail loading the dynamic module "{}". {}'.format(module_name, str(e)))

        return self.modules

    @staticmethod
    def import_type_from_file(script_file: str, type_name: str, predicate=None) -> Type:
        """
        Returns the Type declaration from the specified script file.
        """
        module_name = os.path.splitext(os.path.basename(script_file))[0]

        import importlib.util
        spec = importlib.util.spec_from_file_location(module_name, script_file)
        module_def = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module_def)
        n, t = None, None

        for type_name_i, type_def_i in inspect.getmembers(module_def, predicate):
            n = type_name_i
            t = type_def_i
            if type_name and type_name == n:
                return t

        return t



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import sys
import time
import datetime
import logging
import traceback
import argparse
import json
from typing import List

package_path = os.path.dirname(os.path.dirname(__file__))
if package_path not in sys.path:
    sys.path.insert(0, package_path)

try:
    from geodataflow.core.common import JSONDateTimeEncoder
    from geodataflow.core.processingargs import ProcessingUtils
    from geodataflow.geoext.gdalenv import GdalEnv
    from geodataflow.pipeline.basictypes import AbstractReader, AbstractFilter, AbstractWriter
    from geodataflow.pipeline.pipelinemanager import PipelineManager
except Exception as e:
    raise e


def pipeline_app(command_args: List[str] = None):
    """
    GeodataFlow console application.
    """
    start_time = time.time()

    if not command_args:
        command_args = sys.argv

    # Define the command parameters of the application.
    parser = argparse.ArgumentParser()
    parser.add_argument('--pipeline_file', dest='pipeline_file', required=False, action='store',
                        help='Pipeline file to process.', default='')
    parser.add_argument('--settings_file', dest='settings_file', required=False, action='store',
                        help='Settings file to load (Optional).', default='')
    parser.add_argument('--modules', dest='modules', required=False, action='store_true',
                        help='Show metadata of available Modules and exit without executing the Pipeline.',
                        default='')
    parser.add_argument('--ui_mode', dest='ui_mode', required=False, action='store_true',
                        help='UI mode, the processing task shows a progress bar.')
    #
    parser.add_argument('--temp_path', action='store', required=False,
                        help='Directory to use as temporal folder (Optional).', dest='temp_path', default='')
    parser.add_argument('--log_level', action='store', required=False,
                        help='LOG level to notify application messages (Optional).', dest='log_level',
                        default='INFO')
    parser.add_argument('--log_file', action='store', required=False, help='LOG output file (Optional).',
                        dest='log_file', default='')
    #
    args, pipeline_args = parser.parse_known_args(command_args)

    # Logging application initialization.
    logging_file = None
    if args.pipeline_file:
        logging_file = os.path.splitext(args.pipeline_file)[0] + '.log'
    if args.log_file:
        logging_file = args.log_file
    #
    logging.basicConfig(level=args.log_level, format="[%(levelname)s]: %(message)s")
    #
    if logging_file:
        logging_root = logging.getLogger()
        logging_file_handler = logging.FileHandler(logging_file, mode='w')
        logging_file_handler.setLevel(logging_root.level)
        logging_file_handler.setFormatter(logging_root.handlers[0].formatter)
        logging_root.addHandler(logging_file_handler)

    logging.info('========================================================================')
    logging.info('GeodataFlow:')
    logging.info('Toolkit to run workflows on Geospatial & Earth Observation (EO) data.')
    logging.info('========================================================================')
    logging.info('')

    # Show metadata of available modules and exit?
    if args.modules:
        logging.info('Available Modules...')

        objects = [module_def() for module_def in PipelineManager.modules()]
        objects = sorted(objects, key=lambda obj: obj.className, reverse=False)
        modules_txt = '\n'

        def info_of_module(module):
            """
            Returns info of specified module.
            """
            params_info = ''.join(
                ['\n\t  {}: {}'.format(k, v.get('description')) for k, v in module.params().items()]
            )
            if not params_info:
                params_info = '\n\t  NONE'

            return '  > {}: {}\n\tParams: {}\n'.format(module.className, module.description(), params_info)

        modules_txt += '+ DataSources:\n'

        for obj in [o for o in objects if isinstance(o, (AbstractReader, AbstractWriter))]:
            modules_txt += info_of_module(obj)

        modules_txt += '+ Filters:\n'

        for obj in [o for o in objects if isinstance(o, (AbstractFilter))]:
            modules_txt += info_of_module(obj)

        logging.info(modules_txt)
        logging.warning('The "--modules" flag is present, so exiting...')
        return

    # Initialize the default Settings Manager.
    if not args.settings_file:
        args.settings_file = os.path.splitext(os.path.abspath(__file__))[0] + '.default.settings'

    from geodataflow.core.settingsmanager import Singleton
    app_settings = Singleton.load_from_file(args.settings_file)

    logging.info('Processing file "{}"...'.format(args.pipeline_file))

    # Report context where any module can append its own metadata of results.
    report_context = type('ReportContext', (object,), {})
    report_context.info = {
        'pipelineFile': os.path.basename(args.pipeline_file),
        'startTime': datetime.datetime.now().isoformat(),
        'start': time.time(),
        'status': 'WORKING',
        'message': '',
        'endTime': None,
        'end': None,
        'elapsedTime': None
    }

    # Execute the Workflow.
    with GdalEnv(config_options=GdalEnv.default_options(), temp_path=args.temp_path) as processing_args:
        #
        try:
            if not args.pipeline_file or not os.path.exists(args.pipeline_file):
                raise Exception('Pipeline file not specified!')

            setattr(processing_args, 'ui_mode', ProcessingUtils.strtobool(args.ui_mode))

            # Inject a Dict() as Report context where any module can append its own metadata of results.
            setattr(processing_args, 'reportContext', report_context)

            # Custom modules path.
            custom_modules_path = app_settings.get('GEODATAFLOW__CUSTOM__MODULES__PATH', '')
            custom_modules_path = custom_modules_path.replace('${APP_PATH}', os.path.dirname(__file__))
            custom_modules_path = custom_modules_path.replace('${HOME}', os.path.expanduser('~'))

            # Load & Run workflow.
            pipeline = PipelineManager(config=app_settings, custom_modules_path=custom_modules_path)
            pipeline.load_from_file(args.pipeline_file, pipeline_args)
            pipeline.run(processing_args)

            report_context.info['status'] = 'SUCCESS'
            report_context.info['message'] = 'OK'
        except Exception as e:
            report_context.info['status'] = 'ERROR'
            report_context.info['message'] = str(e)
            logging.error(e)
            traceback.print_exc(file=sys.stdout)
        finally:
            elapsed_time = time.time() - start_time
            elapsed_time = datetime.datetime.utcfromtimestamp(elapsed_time)
            elapsed_text = str(elapsed_time)
            elapsed_text = elapsed_text[elapsed_text.index(' ') + 1:]

            report_context.info['endTime'] = datetime.datetime.now().isoformat()
            report_context.info['end'] = time.time()
            report_context.info['elapsedTime'] = elapsed_text

            if logging_file:
                report_file = os.path.splitext(logging_file)[0] + '.report.json'
                obj = ProcessingUtils.object_as_dict(report_context)
                with open(report_file, mode='w') as fp:
                    json.dump(obj, fp, indent=2, cls=JSONDateTimeEncoder)

            logging.info('--- OK: Process successfully finalized! Elapsed=[{0}]'.format(elapsed_text))
            logging_root = logging.getLogger()
            logging_root.handlers.clear()

    pass


if __name__ == '__main__':
    pipeline_app(command_args=sys.argv[1:])



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import logging
from typing import Dict
from geodataflow.pipeline.filters.EOProductCatalog import EOProductCatalog


class EOProductDataset(EOProductCatalog):
    """
    The Filter extracts Datasets from EO/STAC Collections via spatial & alphanumeric filters.
    """
    def __init__(self):
        EOProductCatalog.__init__(self)
        self.configVars = 'AWS_NO_SIGN_REQUEST=YES'
        self.bands = []
        self.groupByDate = True
        self.clipByAreaOfInterest = True

    def alias(self) -> str:
        """
        Returns the Human alias-name of this Module.
        """
        return 'Dataset'

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Extracts Datasets from EO/STAC Collections via spatial & alphanumeric filters.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'EO STAC Imagery'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        the_params = EOProductCatalog.params(self)
        the_params['configVars'] = {
            'description':
                'Environment variables separated by commas. Commonly used to configure credentials.',
            'dataType': 'string',
            'default': 'AWS_NO_SIGN_REQUEST=YES'
        }
        the_params['bands'] = {
            'description': 'List of Bands to fetch, or a string separated by commas. Empty means fetch all.',
            'dataType': 'string',
            'default': 'B04,B03,B02,B08',
            'placeHolder': 'B04,B03,B02,B08'
        }
        the_params['groupByDate'] = {
            'description': 'Group EO Products by Date.',
            'dataType': 'bool',
            'default': True
        }
        the_params['clipByAreaOfInterest'] = {
            'description': 'Clip EO Products by geometry of input AOI.',
            'dataType': 'bool',
            'default': True
        }
        return the_params

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.core.schemadef import GeometryType
        from geodataflow.geoext.dataset import DATASET_DEFAULT_SCHEMA_DEF

        schema_def = EOProductCatalog.starting_run(self, schema_def, pipeline, processing_args)
        schema_def = schema_def.clone()
        schema_def.geometryType = GeometryType.Polygon
        schema_def.fields = DATASET_DEFAULT_SCHEMA_DEF.copy()
        return schema_def

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        schema_def = self.pipeline_args.schema_def

        from collections import OrderedDict
        from geodataflow.eogeo.dataset import EOGdalDataset
        from geodataflow.geoext.gdalenv import GdalEnv

        eo_config_options = {
            'GDAL_DISABLE_READDIR_ON_OPEN': 'FALSE',
            'CPL_CURL_VERBOSE': 'NO',
            'CPL_DEBUG': 'NO',
            'CPL_VSIL_CURL_ALLOWED_EXTENSIONS': '.tif,.tiff,.vrt,.ovr'
        }
        for item in self.configVars.split(','):
            pair = item.split('=')
            eo_config_options[pair[0].strip()] = pair[1].strip()

        with GdalEnv(config_options=eo_config_options, temp_path=processing_args.temp_data_path()) as eo_env:
            #
            product_groups = OrderedDict()
            bands = list(filter(None, self.bands.replace(' ', '').split(','))) \
                if isinstance(self.bands, str) else self.bands

            # Group EO Products by Date.
            for product in EOProductCatalog.run(self, feature_store, processing_args):
                product_date = product.properties.get('gdf:product_date')

                curr_list = product_groups.get(product_date, [])
                curr_list.append(product)
                product_groups[product_date] = curr_list

            def custom_dataset_op(dataset_ob, operation_args):
                """
                Apply to input GDAL Datasets a custom operation.
                """
                output_crs = operation_args.get('crs')
                clip_geometry = operation_args.get('clippingGeom', None)

                # Figure out the best PixelSize of output Dataset.
                best_datasets = [
                    d for d in operation_args.get('datasets', [])
                    if d.get_spatial_srid() == output_crs.to_epsg()
                ]
                if best_datasets:
                    best_dataset = best_datasets[0]
                    best_info = best_dataset.get_metadata()
                    output_res_x = best_info.get('pixelSizeX')
                    output_res_y = best_info.get('pixelSizeY')

                    # Clamping Geometry to output PixelSize.
                    clip_geometry = best_dataset.clamp_geometry(clip_geometry)
                else:
                    output_res_x = None
                    output_res_y = None

                # Warp & Clip Dataset.
                return dataset_ob.warp(output_crs, output_res_x, output_res_y, clip_geometry)

            logging.info('Fetching EO Products as GDAL Datasets...')

            # Fetch EO Products as GDAL Datasets.
            for product_date, products in product_groups.items():
                logging.info('Starting conversion of EO Product, Date: {}...'.format(product_date))

                clipping_geom = products[0].areaOfInterest.geometry \
                    if self.clipByAreaOfInterest and products[0].areaOfInterest else None

                custom_args = {
                    'crs': schema_def.crs,
                    'clippingGeom': clipping_geom
                }

                # Convert to a mosaic of GDAL Datasets.
                if self.groupByDate:
                    datasets = [EOGdalDataset.open([product.assets for product in products],
                                                   bands,
                                                   eo_env,
                                                   custom_dataset_op,
                                                   custom_args)]
                else:
                    datasets = [EOGdalDataset.open(product.assets,
                                                   bands,
                                                   eo_env,
                                                   custom_dataset_op,
                                                   custom_args) for product in products]

                logging.info('Done!')

                for dataset in datasets:
                    dataset.user_data['areaOfInterest'] = products[0].areaOfInterest
                    dataset.user_data['product_type'] = self.product
                    dataset.user_data['product_date'] = product_date
                    yield dataset
            #
        pass



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import logging
from typing import Dict

from geodataflow.core.capabilities import StoreCapabilities
from geodataflow.pipeline.basictypes import AbstractWriter


class FeatureWriter(AbstractWriter):
    """
    Writes Features with Geometries to a Geospatial DataStore using OGR providers.
    """
    def __init__(self):
        AbstractWriter.__init__(self)
        self._featureStore = None
        self.connectionString = ''

    def description(self) -> str:
        """
        Returns the Description text of this Module.
        """
        return 'Writes Features with Geometries to a Geospatial DataStore using OGR providers.'

    def category(self) -> str:
        """
        Returns the category or group to which this Module belongs.
        """
        return 'Output'

    def params(self) -> Dict:
        """
        Returns the declaration of parameters supported by this Module.
        """
        return {
            'connectionString': {
                'description':
                    "Connection string of the FeatureStore ('.geojson', '.gpkg', '.shp.zip' are supported).",
                'dataType': 'string',
                'default': 'output.gpkg',
                'extensions': ['.geojson', '.gpkg', '.shp.zip']
            }
        }

    def test_capability(self, connection_string: str, capability: StoreCapabilities) -> bool:
        """
        Returns this Module supports the specified ConnectionString and named StoreCapability.
        """
        from geodataflow.geoext.commonutils import GdalUtils
        return GdalUtils.test_ogr_capability(connection_string, capability)

    def starting_run(self, schema_def, pipeline, processing_args):
        """
        Starting a new Workflow on Geospatial data.
        """
        from geodataflow.geoext.featurestore import OgrFeatureStore
        self._featureStore = OgrFeatureStore()

        # Create the FeatureStore.
        schema_def = self._featureStore.create(self.connectionString, schema_def)
        return schema_def

    def run(self, feature_store, processing_args):
        """
        Transform input Geospatial data. It should return a new iterable set of Geospatial features.
        """
        if self._featureStore:
            feature_count = 0

            # Group n features per transaction (default 20000).
            # Increase the value for better performance when writing into DBMS drivers that have transaction
            # support.
            # n can be set to unlimited to load the data into a single transaction.
            # https://github.com/OSGeo/gdal/blob/release/3.2/gdal/swig/python/samples/ogr2ogr.py#L1540
            # https://gdal.org/programs/ogr2ogr.html#cmdoption-ogr2ogr-gt
            cache_size = self.cacheSize if hasattr(self, 'cacheSize') else 20000

            for feature in self._featureStore.write_features(features=feature_store, cache_size=cache_size):
                feature_count += 1
                yield feature

            logging.info('{:,} Features saved to "{}".'.format(feature_count, self.connectionString))
            return

        raise Exception('FeatureStore is not open!')

    def finished_run(self, pipeline, processing_args):
        """
        Finishing a Workflow on Geospatial data.
        """
        self._featureStore = None
        return True



# -*- coding: utf-8 -*-
"""
===============================================================================

   GeodataFlow:
   Toolkit to run workflows on Geospatial & Earth Observation (EO) data.

   Copyright (c) 2022, Alvaro Huarte. All rights reserved.

   Redistribution and use of this code in source and binary forms, with
   or without modification, are permitted provided that the following
   conditions are met:
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
   TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
   OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SAMPLE CODE, EVEN IF
   ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

===============================================================================
"""

import os
import uuid
import logging
from typing import Any, Dict, Iterable, List, Tuple, Union

from geodataflow.geoext.commonutils import GeometryUtils
from geodataflow.geoext.gdalenv import GdalEnv
from geodataflow.core.schemadef import DataType, FieldDef

from osgeo import gdalconst as gdal_const
from osgeo import gdal_array
from osgeo import osr
osr.UseExceptions()

# Default schema of a GDAL Dataset.
DATASET_DEFAULT_SCHEMA_DEF = [
    FieldDef(name='srid', data_type=DataType.Integer),
    FieldDef(name='dataType', data_type=DataType.Integer),
    FieldDef(name='rasterSizeX', data_type=DataType.Integer),
    FieldDef(name='rasterSizeY', data_type=DataType.Integer),
    FieldDef(name='rasterCount', data_type=DataType.Integer),
    FieldDef(name='pixelSizeX', data_type=DataType.Float),
    FieldDef(name='pixelSizeY', data_type=DataType.Float),
    FieldDef(name='noData', data_type=DataType.Float)
]


class GdalDataset:
    """
    Wrapper class of the GDAL Dataset object.
    """
    def __init__(self,
                 dataset_or_path: Union["GdalDataset", str],
                 gdal_env: GdalEnv = None,
                 user_data: Dict = {},
                 recyclable: bool = False):

        self.type = 'Raster'
        self.user_data = user_data or {}
        self._gdal_env = gdal_env
        self._dataset_path = None
        self._recyclable = recyclable
        self._init_object(dataset_or_path)

    def __del__(self):
        self._gdal_env = None
        self._dataset_path = None
        self._recyclable = False
        self._dataset = None

    def _init_object(self, dataset_ref: Union["GdalDataset", str]) -> None:
        """
        Helper method to initialize some GDAL attributes.
        """
        if isinstance(dataset_ref, GdalDataset):
            self._gdal_env = dataset_ref._gdal_env
            self._dataset_path = dataset_ref._dataset_path
            self._dataset = dataset_ref._dataset
        elif isinstance(dataset_ref, str):
            gdal = self.env().gdal()
            self._dataset_path = dataset_ref
            self._dataset = gdal.Open(dataset_ref, gdal.GA_ReadOnly)
        else:
            self._dataset_path = None
            self._dataset = dataset_ref

        rs_band = self._dataset.GetRasterBand(1)
        self.gdal_type = rs_band.DataType
        self.data_type = gdal_array.GDALTypeCodeToNumericTypeCode(self.gdal_type)
        rs_band = None

    @property
    def geometry(self):
        """
        Returns the Geometry of this Dataset.
        """
        envelope = self.get_envelope()
        geometry = GeometryUtils.create_geometry_from_bbox(*envelope)
        geometry = geometry.with_srid(self.get_spatial_srid())
        return geometry

    @property
    def properties(self) -> Dict[str, Any]:
        """
        Returns the Properties of this Dataset.
        """
        info = self.get_metadata()
        return {field_def.name: info.get(field_def.name) for field_def in DATASET_DEFAULT_SCHEMA_DEF}

    def env(self) -> GdalEnv:
        """
        Returns the related GdalEnv instance of this Dataset.
        """
        return self._gdal_env if self._gdal_env else GdalEnv.default()

    def dataset(self):
        """
        Returns the GDAL Dataset managed.
        """
        return self._dataset

    def dataset_path(self, force_exists: bool = False) -> str:
        """
        Returns the path of this Dataset.
        """
        if not self._dataset_path and force_exists:
            gdal_env = self.env()
            gdal = gdal_env.gdal()
            temp_file = os.path.join(gdal_env.temp_data_path(), str(uuid.uuid4()).replace('-', '') + '.tif')

            gdal.Translate(temp_file,
                           self._dataset,
                           creationOptions=['TILED=YES', 'COMPRESS=DEFLATE', 'PREDICTOR=2'])

            self._dataset_path = temp_file
            self._recyclable = True

        return self._dataset_path

    def recycle(self) -> bool:
        """
        Recycle temporary file resources created by this Dataset.
        """
        if self._recyclable and self._dataset_path and os.path.exists(self._dataset_path):
            self._dataset = None
            os.remove(self._dataset_path)
            self._dataset_path = None
            self._recyclable = False
            return True

        return False

    def get_metadata(self) -> Dict[str, Any]:
        """
        Returns the main geospatial metadata of this Dataset.
        """
        rs_band = self._dataset.GetRasterBand(1)
        no_data = rs_band.GetNoDataValue() if rs_band.GetNoDataValue() is not None else 0.0
        rs_band = None

        envelope = self.get_envelope()
        pixel_size_x = (envelope[2] - envelope[0]) / float(self._dataset.RasterXSize)
        pixel_size_y = (envelope[3] - envelope[1]) / float(self._dataset.RasterYSize)

        info = {
            'srid': self.get_spatial_srid(),
            'envelope': envelope,
            'dataType': self.gdal_type,
            'rasterSizeX': self._dataset.RasterXSize,
            'rasterSizeY': self._dataset.RasterYSize,
            'rasterCount': self._dataset.RasterCount,
            'pixelSizeX': pixel_size_x,
            'pixelSizeY': pixel_size_y,
            'noData': no_data
        }
        return info

    def get_envelope(self) -> Tuple[float, float, float, float]:
        """
        Returns the spatial envelope of this Dataset.
        """
        geo_transform = self._dataset.GetGeoTransform()
        c = geo_transform[0]
        a = geo_transform[1]
        b = geo_transform[2]
        f = geo_transform[3]
        d = geo_transform[4]
        e = geo_transform[5]
        t = 0  # Texel offset, by default the texel is centered to CENTER-CENTER pixel.
        col = 0
        row = 0
        env_a = [a * (col + t) + b * (row + t) + c, d * (col + t) + e * (row + t) + f]
        col = self._dataset.RasterXSize
        row = self._dataset.RasterYSize
        env_b = [a * (col + t) + b * (row + t) + c, d * (col + t) + e * (row + t) + f]
        min_x = min(env_a[0], env_b[0])
        min_y = min(env_a[1], env_b[1])
        max_x = max(env_a[0], env_b[0])
        max_y = max(env_a[1], env_b[1])
        return min_x, min_y, max_x, max_y

    def get_spatial_srid(self) -> int:
        """
        Returns the EPSG SRID code of the this Dataset.
        """
        spatial_wkt = self._dataset.GetProjection()
        spatial_ref = osr.SpatialReference()
        spatial_ref.ImportFromWkt(spatial_wkt)

        if spatial_ref is None:
            return 0

        srid = 0
        try:
            srid = spatial_ref.GetAuthorityCode(None)

            if srid is None or srid == 0 or srid == '':
                temp = spatial_ref.AutoIdentifyEPSG()
                if temp is not None and temp != 0 and temp != '':
                    srid = int(temp)
            else:
                srid = int(srid)

        except Exception as e:
            logging.warning('Error reading the SRID of the CRS="{}" Cause="{}"'.format(str(spatial_ref), e))

        return srid

    def clamp_geometry(self, geometry, padding_val: int = 0, minimum_ts: int = 0):
        """
        Clamp the specified Envelope/Geometry aligned to current GRID of this Dataset.
        """
        if not geometry:
            return geometry

        info = self.get_metadata()
        is_envelope = False

        if isinstance(geometry, list):
            geometry = GeometryUtils.create_geometry_from_bbox(*geometry)
            is_envelope = True
        else:
            transform_fn = GeometryUtils.create_transform_function(geometry.get_srid(), info.get('srid'))
            geometry = transform_fn(geometry)

        pixel_size_x = info.get('pixelSizeX')
        pixel_size_y = info.get('pixelSizeY')

        envelope = GeometryUtils.clamp_envelope(
            envelope=list(geometry.bounds),
            pixel_size_x=pixel_size_x,
            pixel_size_y=pixel_size_y,
            padding_val=padding_val,
            minimum_ts=minimum_ts
        )
        if is_envelope:
            return envelope

        geometry = GeometryUtils.create_geometry_from_bbox(*envelope)
        geometry = geometry.with_srid(info.get('srid'))
        return geometry

    def warp(self,
             output_crs=None, output_res_x: float = None, output_res_y: float = None, output_geom=None,
             resample_arg: int = gdal_const.GRA_Bilinear) -> "GdalDataset":
        """
        Warp this Dataset by specified parameters.
        """
        gdal_env = self.env()
        gdal = gdal_env.gdal()
        info = self.get_metadata()

        if isinstance(output_geom, list):
            output_geom = GeometryUtils.create_geometry_from_bbox(*output_geom)
            output_geom = output_geom.with_srid(GeometryUtils.get_srid(output_crs) if output_crs else info.get('srid'))

        if output_geom:
            source_crs = GeometryUtils.get_spatial_crs(info.get('srid'))
            output_crs = GeometryUtils.get_spatial_crs(output_crs) if output_crs else source_crs
            window_crs = GeometryUtils.get_spatial_crs(GeometryUtils.get_srid(output_geom))

            # Calculate intersection of two Envelopes.
            if source_crs and output_crs:
                transform_fn = GeometryUtils.create_transform_function(source_crs, output_crs)
                geometry_a = GeometryUtils.create_geometry_from_bbox(*self.get_envelope())
                geometry_a = transform_fn(geometry_a)
            else:
                geometry_a = GeometryUtils.create_geometry_from_bbox(*self.get_envelope())

            if window_crs and output_crs:
                transform_fn = GeometryUtils.create_transform_function(window_crs, output_crs)
                geometry_b = transform_fn(output_geom)
            else:
                geometry_b = output_geom

            # Run intersection!
            geometry = geometry_a.intersection(geometry_b).with_srid(output_crs.to_epsg())

            if output_res_x and output_res_y:
                envelope = GeometryUtils.clamp_envelope(list(geometry.bounds), output_res_x, output_res_y)
            else:
                envelope = list(geometry.bounds)

            # Define the Warping options for resampling and reprojection.
            options = gdal.WarpOptions(format='Gtiff',
                                       outputBounds=envelope,
                                       resampleAlg=resample_arg,
                                       xRes=output_res_x,
                                       yRes=output_res_y,
                                       srcSRS='EPSG:' + str(source_crs.to_epsg()),
                                       srcNodata=info.get('noData'),
                                       dstSRS='EPSG:' + str(output_crs.to_epsg()),
                                       dstNodata=info.get('noData'),
                                       copyMetadata=True,
                                       creationOptions=['TILED=YES', 'COMPRESS=DEFLATE', 'PREDICTOR=2'])
        else:
            source_srid = info.get('srid')
            output_srid = GeometryUtils.get_srid(output_crs) if output_crs else source_srid

            # Define the Warping options for resampling and reprojection.
            options = gdal.WarpOptions(format='Gtiff',
                                       resampleAlg=resample_arg,
                                       xRes=output_res_x,
                                       yRes=output_res_y,
                                       srcSRS='EPSG:' + str(source_srid),
                                       srcNodata=info.get('noData'),
                                       dstSRS='EPSG:' + str(output_srid),
                                       dstNodata=info.get('noData'),
                                       copyMetadata=True,
                                       creationOptions=['TILED=YES', 'COMPRESS=DEFLATE', 'PREDICTOR=2'])

        warp_name = str(uuid.uuid1()).replace('-', '')
        warp_file = os.path.join(gdal_env.temp_data_path(), 'temp_WARP_{}.tif'.format(warp_name))
        dataset = gdal.Warp(warp_file, self._dataset, options=options)
        if dataset:
            dataset.FlushCache()
            dataset = None

        return GdalDataset(warp_file, gdal_env, self.user_data.copy(), True)

    def calc(self,
             band_names: Iterable[str], band_expression: str, no_data: float = -9999.0) -> "GdalDataset":
        """
        Apply the specified Raster Calc expression (numpy syntax) to this Dataset.
        """
        letters = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
                   'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'Y', 'Z']

        band_names = band_names or self.user_data.get('bands')
        if not band_names:
            raise Exception('BandNames of GdalDataset::apply_calc() method can not be empty!')

        dataset_file = self.dataset_path(force_exists=True)
        gdal_env = self.env()
        no_data = no_data if no_data is not None else -9999.0

        calc_name = str(uuid.uuid1()).replace('-', '')
        calc_file = os.path.join(gdal_env.temp_data_path(), 'temp_CALC_{}.tif'.format(calc_name))
        calc_args = ''
        calc_expr = band_expression

        for i, band_name in enumerate(band_names):
            if band_name in calc_expr:
                calc_args += ' -{} "{}" --{}_band={}'.format(letters[i], dataset_file, letters[i], i + 1)
                calc_expr = calc_expr.replace(band_name, letters[i] + '.astype(numpy.float32)')

        for calc_item in calc_expr.split(','):
            calc_args += ' --calc="{}"'.format(calc_item)

        # Perform GdalCalc to this GDAL Dataset.
        scripts_path = GdalEnv.gdal_scripts_path()
        command_line = 'python "' + os.path.join(scripts_path, 'gdal_calc.py') + '"'
        command_line += ' --overwrite --type=Float32 --NoDataValue={} --quiet'.format(no_data)
        command_line += calc_args
        command_line += ' --outfile="' + calc_file + '"'
        command_line += ' --format Gtiff --co "TILED=YES" --co "COMPRESS=DEFLATE" --co "PREDICTOR=2"'
        gdal_env.run(command_line)
        #
        return GdalDataset(calc_file, gdal_env, self.user_data.copy(), True)

    def split(self, tile_size_x: int, tile_size_y: int, padding: int = 0) -> Iterable["GdalDataset"]:
        """
        Split this Dataset into tiles.
        """
        gdal_env = self.env()
        driver = gdal_env.gdal().GetDriverByName('Gtiff')
        format_options = ['TILED=YES', 'COMPRESS=DEFLATE', 'PREDICTOR=2']

        geo_transform = self._dataset.GetGeoTransform()
        raster_size_x = self._dataset.RasterXSize
        raster_size_y = self._dataset.RasterYSize
        raster_count = self._dataset.RasterCount
        spatial_wkt = self._dataset.GetProjection()

        x, y, w, h = 0, 0, 0, 0
        item_count = len(range(0, raster_size_x, tile_size_x)) * len(range(0, raster_size_y, tile_size_y))
        item_index = 0

        for i in range(0, raster_size_x, tile_size_x):
            y = 0
            for j in range(0, raster_size_y, tile_size_y):
                w = min(i + tile_size_x, raster_size_x) - i
                h = min(j + tile_size_y, raster_size_y) - j

                if padding:
                    l_ = max(0, x - padding)
                    t_ = max(0, y - padding)
                    r_ = min(x + w + padding, raster_size_x)
                    b_ = min(y + h + padding, raster_size_y)
                else:
                    l_, t_, r_, b_ = x, y, x + w, y + h

                x_pos = geo_transform[0] + (l_ * geo_transform[1])
                y_pos = geo_transform[3] + (t_ * geo_transform[5])
                w = r_ - l_
                h = b_ - t_

                logging.info(
                    'Splitting Dataset, XY=({} {}), RECT=({} {} {} {}), Step=({} of {})...'
                    .format(x_pos, y_pos, l_, t_, r_, b_, item_index+1, item_count))

                # Create child GDAL Dataset if current Tile.
                split_name = str(uuid.uuid1()).replace('-', '')
                split_file = os.path.join(gdal_env.temp_data_path(), 'temp_SPLIT_C_{}.tif'.format(split_name))
                output = driver.Create(split_file, w, h, raster_count, self.gdal_type, options=format_options)
                output.SetGeoTransform([x_pos, geo_transform[1], 0.0, y_pos, 0.0, geo_transform[5]])
                output.SetProjection(spatial_wkt)
                metadata = self._dataset.GetMetadata()
                if metadata is not None:
                    output.SetMetadata(metadata)

                # Fill data of Bands.
                for band_index in range(0, raster_count):
                    band_s = self._dataset.GetRasterBand(band_index + 1)
                    raster = band_s.ReadAsArray(xoff=l_, yoff=t_, win_xsize=w, win_ysize=h)
                    nodata = band_s.GetNoDataValue()
                    band_s = None

                    band_t = output.GetRasterBand(band_index + 1)
                    band_t.Fill(nodata)
                    band_t.SetNoDataValue(nodata)
                    band_t.WriteArray(raster)
                    band_t.FlushCache()
                    band_t = None

                output.FlushCache()
                item_index += 1

                # print('{} -> {}'.format([x, y, w, h], [l, t, r, b]))
                yield GdalDataset(output, gdal_env, self.user_data.copy(), True)
                y += tile_size_y

            x += tile_size_x

        pass

    @staticmethod
    def mosaic_of_datasets(datasets: List["GdalDataset"],
                           resample_arg: int = gdal_const.GRA_Bilinear) -> "GdalDataset":
        """
        Returns a mosaic of the specified Dataset collection.
        """
        if len(datasets) == 0:
            return None
        if len(datasets) == 1:
            return datasets[0]

        first_info = datasets[0].get_metadata()
        user_data = {}
        mosaic_of_files = list()

        gdal_env = datasets[0].env()
        gdal = gdal_env.gdal()
        gdal_config = gdal_env.gdal_config()

        # Do we need transform input GDAL Datasets, otherwise "gdalbuildvrt" will fail.
        for dataset in datasets:
            info = dataset.get_metadata()
            user_data.update(dataset.user_data)

            if info.get('srid') == first_info.get('srid'):
                mosaic_of_files.append(dataset.dataset_path(force_exists=True))
            else:
                transform_fn = GeometryUtils.create_transform_function(
                    info.get('srid'), first_info.get('srid')
                )

                geometry = GeometryUtils.create_geometry_from_bbox(*dataset.get_envelope())
                geometry = transform_fn(geometry)
                envelope = list(geometry.bounds)
                envelope = GeometryUtils.clamp_envelope(
                    envelope,
                    first_info.get('pixelSizeX'),
                    first_info.get('pixelSizeY')
                )

                # Define the Warping options for resampling and reprojection.
                options = gdal.WarpOptions(format='Gtiff',
                                           outputBounds=envelope,
                                           resampleAlg=resample_arg,
                                           xRes=first_info.get('pixelSizeX'),
                                           yRes=first_info.get('pixelSizeY'),
                                           srcSRS='EPSG:' + str(info.get('srid')),
                                           srcNodata=info.get('noData'),
                                           dstSRS='EPSG:' + str(first_info.get('srid')),
                                           dstNodata=first_info.get('noData'),
                                           copyMetadata=True,
                                           creationOptions=['TILED=YES', 'COMPRESS=DEFLATE', 'PREDICTOR=2'])

                warp_name = str(uuid.uuid1()).replace('-', '')
                warp_file = os.path.join(gdal_env.temp_data_path(), 'temp_WARP_C_{}.tiff'.format(warp_name))
                dataset = gdal.Warp(warp_file, dataset.dataset(), options=options)
                if dataset:
                    dataset.FlushCache()
                    dataset = None

                mosaic_of_files.append(warp_file)
            #

        # Merging all files to one unique one.
        mosaic_name = str(uuid.uuid1()).replace('-', '')
        mosaic_file = os.path.join(gdal_env.temp_data_path(), 'temp_MOSAIC_{}.vrt'.format(mosaic_name))
        command_line = 'gdalbuildvrt {} "{}" {}'.format(gdal_config, mosaic_file, ' '.join(mosaic_of_files))
        gdal_env.run(command_line)
        #
        return GdalDataset(mosaic_file, gdal_env, user_data)
